/*
 *  Author: Anthony Reese & David Huynh
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 */

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;

namespace rhodes.graphics
{
    /*! Author::  Anthony Reese & David Huynh
     * Handles the current view of a 2D environment. Can be moved, scaled, and rotated.
     * The camera has an ideal view of the sprites it is tracking. */
    public class Camera2D {
        private struct TargetInfo {
            public Sprite.PositionChangedDelegate _delegate;
            public Vector2 _derivedPosition;
            public short _width;
            public short _height;
        };
        
        private static volatile Camera2D _instance;
        private static object _syncRoot = new Object();
        
        public static float MAXIMUM_ZOOM = 0.9f;
        public static float MINIMUM_ZOOM = 0.3f;

        private Vector2 _view;
        private Vector2 _dimension;
        private float _zoom; //!< how much the camera scaled
        private Vector2 _idealCenter; //!< where the camera wants to be centered
        private float _idealZoom; //!< the zoom level the camera wants to be
		Rectangle _world;

        private SortedDictionary<String, TargetInfo> _targets;

        /*! Creates a empty camera */
        public Camera2D()
        {
            _view = Vector2.Zero;
            _zoom = 1;
            _idealZoom = 1;
            _idealCenter = Vector2.Zero;
            _targets = new SortedDictionary<String, TargetInfo>();
			_world = new Rectangle(-503, -303, 1903, 1403);//1400, 1100)
        }

        /*! Creates a default camera that is set at (0,0) with a scren size of 
         * screenWdithxscreenHeight.
         * \param screenWidth The screen width 
         * \param screenHeight The screen height */
        public void initialize(int screenWidth, int screenHeight) {
            _dimension = new Vector2(screenWidth, screenHeight);
            
        }

        /*! Returns singleton instance of the Camera2D */
        public static Camera2D getSingleton {
            get {
                if(_instance == null) {
                    lock(_syncRoot) {
                        if(_instance == null)
                            _instance = new Camera2D();
                    }
                }
                return _instance;
            }
        }

        /*! Moves the camera by the positionSpeed
         * \param moveVector Vector used to translate the view position */
        public void Translate(Vector2 moveVector) {
            _view = _view + moveVector;
        }

        /*! Moves the camera by the scaleSpeed
         * \param scaleValue The scale value used to modified the zoom in the camera */
        public void Scale(float scaleValue) {
            _zoom += scaleValue;
        }

        
        /*! Tells the camera what sprites to keep an "eye" on
         * \param spriteName The sprite to be registered to the camera */
        public bool register(String spriteName) {
            if(!_targets.ContainsKey(spriteName)) {
                Sprite sprite = SpriteManager.getSingleton.get(spriteName);
                TargetInfo info = new TargetInfo();

                info._delegate = new Sprite.PositionChangedDelegate(notifyPosition);
                info._derivedPosition = sprite.Position;
                info._width = sprite.Width;
                info._height = sprite.Height;

                _targets.Add(spriteName, info);
                sprite.PositionChangedEvent += info._delegate;

                return true;
            }

            return false;
        }

        /*! Removes a target that the camera is tracking.
         * \param spriteName The sprite to be unregistered to the camera */
        public bool unregister(String spriteName) {
            if(_targets.ContainsKey(spriteName)) {
                Sprite sprite = SpriteManager.getSingleton.get(spriteName);

                sprite.PositionChangedEvent -= _targets[spriteName]._delegate;
                _targets.Remove(spriteName);

                return true;
            }

            return false;
        }

        /*! Called by sprites when a position has changed.
         * \param name The name of the sprite
         * \param sprite The sprite */
        public void notifyPosition(String spriteName, Sprite sprite) {
            if(_targets.ContainsKey(spriteName)) {
                TargetInfo info = _targets[spriteName];
                info._derivedPosition = sprite.Position;
				_targets[spriteName] = info;
            }
        }

        /*! Updates the camera based on the ideal view
         * \param elapsedTime The time elapsed since last tick in milliseconds. */
        public void update(float elapsedTime) {
            if(_targets.Count > 0) {
                setUpIdeal();

                // Hey Anthony, you should probably do things with respect to time
                // Also the speeds should be variable.
                Vector2 move = (_idealCenter - _view) * elapsedTime * 5.0f;
                Translate(move);
                Scale((_idealZoom - _zoom) * elapsedTime * 5.0f);
            }
        }

        /*! Sets up the ideal view, scale, and rotation fo where the camera wants to be. */
        public void setUpIdeal() 
        {
            //set inital vales for the min and max values
            SortedDictionary<String,TargetInfo>.Enumerator iter = _targets.GetEnumerator();
            iter.MoveNext();
            TargetInfo first = iter.Current.Value;
            Vector2 min = first._derivedPosition - new Vector2(first._width, first._height);
            Vector2 max = first._derivedPosition + new Vector2(first._width, first._height);
            
            // computer ideal position by creating a box around  the observable objects
            foreach(TargetInfo target in _targets.Values) {
         
                min = Vector2.Min(min, target._derivedPosition -
                    new Vector2(target._width, target._height) );
                max = Vector2.Max(max, target._derivedPosition + 
                    new Vector2(target._width, target._height) );
            }
            
            // when the camera's center, moves to the ideal center, this change must be 
            // relected on the positon
            Vector2 idealDimension = max - min;

            // compute ideal scale
            _idealZoom = (float)Math.Min(
                _dimension.X / idealDimension.X, 
                _dimension.Y / idealDimension.Y );

            // don't zoom if the screen is at the maximum zoom
            if (_idealZoom >= MAXIMUM_ZOOM)
                _idealZoom = MAXIMUM_ZOOM;

            if (_idealZoom <= MINIMUM_ZOOM)
                _idealZoom = MINIMUM_ZOOM;

            _idealCenter = min + (max - min) / 2 - (_dimension / 2) / _idealZoom;
        }

        /*! Properties */
        public Vector2 Position { get { return _view; } set { _view = value; } }
        public float Zoom { get { return _zoom; } set { _zoom = value; } }
        public Vector2 Dimension { get { return _dimension; } set { _dimension = value; } }
    }
}