/*
 *  Author: Brian Murphy
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 */

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using rhodes.graphics.particles;
using rhodes.graphics.particles.emitters;

namespace rhodes.graphics
{
	public class ParticleManager
	{

        // Objects to ensure singleton design
        private static volatile ParticleManager _instance;
        private static object _syncRoot = new Object();

		private static int _nameNum = 0;
        private SortedDictionary<String, ParticleEngine> _particleEngines;

		/*! Constructor */
        private ParticleManager()
		{
            _particleEngines = new SortedDictionary<string, ParticleEngine>();
        }

        /*! Returns singleton instance of the SceneManager */
        public static ParticleManager getSingleton
		{
            get {
                if (_instance == null) {
                    lock (_syncRoot) {
                        if (_instance == null)
                            _instance = new ParticleManager();
                    }
                }

                return _instance;
            }
        }

        /*! Loads and returns a particle engine 
         * \param particleEngineName The name of the particle engine
         * \param particleEngineFilename The path and filename of the particle engine 
         * \return Returns a particle engine */
        public ParticleEngine load(string particleEngineName, string particleEngineFilename) {
            //Console.Out.WriteLine("loaded " + particleEngineName);
            ParticleEngine retVal = null;
            if (_particleEngines.ContainsKey(particleEngineName)) {
                //Console.Error.WriteLine("Particle entity: " + particleEngineName + " already exists.");
                clone(_particleEngines[particleEngineName], ref retVal);
				_particleEngines.Add(particleEngineName + _nameNum, retVal);
            } else {
                retVal = GraphicsCore.getSingleton.Content.Load<ParticleEngine>(particleEngineFilename);
                if (retVal != null) {
                    _particleEngines.Add(particleEngineName, retVal);
                }
            }
			_nameNum++;
            return retVal;
        }

		/*! Makes a copy of the given particle engine
		* \param pEngine The particle engine to copy
		* \param particle The particle engine to set*/
		private void clone(ParticleEngine pEngine, ref ParticleEngine particle)
		{
			List<Emitter> emitters = new List<Emitter>();
			PointEmitter point = null;
			SplineEmitter spline = null;
			DirectionalEmitter dir = null;
			SprayEmitter spray = null;
			foreach (Emitter emitter in pEngine.Emitters)
			{
				if (emitter is PointEmitter)
				{
					((PointEmitter)emitter).clone(ref point);
					emitters.Add(point);
				}
				else if (emitter is SplineEmitter)
				{
					((SplineEmitter)emitter).clone(ref spline);
					emitters.Add(spline);
				}
				else if (emitter is DirectionalEmitter)
				{
					((DirectionalEmitter)emitter).clone(ref dir);
					emitters.Add(dir);
				}
				else
				{
					((SprayEmitter)emitter).clone(ref spray);
					emitters.Add(spray);
				}
			}
			particle = new ParticleEngine(emitters);
		}

        /*! Updates all paticle engines in the manager.
         * \param elapsedTime The time elapsed since last tick in milliseconds. */
        public void update(float elapsedTime) {
            foreach (ParticleEngine engine in _particleEngines.Values) {
                engine.Update(elapsedTime);
            }
        }

        /*! Gets a particle engine by name.
         * \param name The name of the particle engine.
         * \return Returns a particle engine. */
        public ParticleEngine get(String name) {
            if (_particleEngines.ContainsKey(name)) {
                return _particleEngines[name];
            }

            return null;
        }

        /*! Removes all particle engines */
        public void removeAll() {
            _particleEngines.Clear();
        }

	}
}
