/*
 *  Author: Josh Wilson and David Huynh and Robert Yates
 *  Copyright 2007. All rights reserved.
 */

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;

using FarseerGames.FarseerPhysics;
using FarseerGames.FarseerPhysics.Dynamics;
using FarseerGames.FarseerPhysics.Collisions;

using RhodesProcessorLib;
using SedohrProcessorLib;
using rhodes.graphics;
using rhodes.engine;
using rhodes.physics;
using rhodes.audio;
using rhodes.input;

using sedohr.player;
using sedohr.weapon;
using sedohr.objects;

namespace sedohr
{
	/// <summary>
	/// Stores information about the demo "screen" being displayed in the program. 
	/// </summary>
	class AnthonyGameState : GameState
	{
		/// <summary>
		/// Contains the name of the game state
		/// </summary>
		public readonly static string GAME_STATE_NAME = "anthony";

		Level level;

		/// <summary>
		/// Contains the PlayerManager that controls all the players in the game state
		/// </summary>
		PlayerManager playerManager;

		/// <summary>
		/// Contains the mouse cursor sprite
		/// </summary>
		Sprite mouseCursor;

		/// <summary>
		/// Contains the background sprite
		/// </summary>
		Sprite background;

		/// <summary>
		/// Contains the input used by the player
		/// </summary>
		Input input;

		/// <summary>
		/// Contains the light in the scene
		/// </summary>
		Light light1;

		ObjectManager objectManager;
		/// <summary>
		/// Contains a list of all sprite items in the HUD
		/// </summary>
		private SortedDictionary<String, Sprite> _hudItems;

		/// <summary>
		/// Contains a list of all message items in the hud
		/// </summary>
		private SortedDictionary<String, Message> _hudMsgItems;

		/// <summary>
		/// Creates the demo game state with a given GameStateManager.
		/// </summary>
		/// <param name="parent">The GameStateManager that is storing this game state.</param>
		public AnthonyGameState(GameStateManager parent)
			: base(parent)
		{
			input = parent.getInput();
			_hudItems = new SortedDictionary<string, Sprite>();
			_hudMsgItems = new SortedDictionary<string, Message>();
			objectManager = new ObjectManager();
		}

		/// <summary>
		/// Loads all the assets for this state.
		/// </summary>
		public override void load()
		{
			Level level = GraphicsCore.getSingleton.Content.Load<Level>("assets/levels/space");
			level.load(GAME_STATE_NAME, objectManager);

			Layer mainLayer = SceneManager.getSingleton.getActiveScene().getLayer(5);

			//Load the players and their weapons
			playerManager = new PlayerManager(input, mainLayer, objectManager);
			ProjectileManager.getSingleton.Players = playerManager.Players;

			//light1 = new Light();
			//light1.Position = new Vector3(0, 300, 10);
			//light1.IsAttenuated = false;
			//mainLayer.AddLight("mainLight", light1);
			//Load HUD
			loadHUD();

			//Load the crosshair

			for (int i = 1; i <= playerManager.Players.Count; i++)
			{
				Camera2D.getSingleton.unregister("player" + i);
			}
			Camera2D.getSingleton.Position = new Vector2(-550, -350);
			Camera2D.getSingleton.Zoom = 0.4f;

			mouseCursor = SpriteManager.getSingleton.load("mouseCursor", "assets/sprites/crosshair");
			mainLayer.AddSprite("mouseCursor", mouseCursor);
		}

		/// <summary>
		/// Loads the HUD for the Game
		/// </summary>
		private void loadHUD()
		{
			Sprite sprite;
			Message msg;

			for (int i = 1; i <= playerManager.Players.Count; i++)
			{
				//NOTE*** Create one "avatar" to contain the player avatar image box, weapon image box, healthbar box, and kill/death box
				sprite = GuiManager.getSingleton.load("player" + i + "avatar", "assets/sprites/player" + i + "avatar");
				sprite.Position = new Vector2((i - 1) * 200 + 60, 20);
				sprite.pause();
				_hudItems.Add("player" + i + "avatar", sprite);

				sprite = GuiManager.getSingleton.load("player" + i + "health", "assets/sprites/healthbartest");
				sprite.Position = new Vector2((i - 1) * 200 + 128, 10);
				sprite.pause();
				_hudItems.Add("player" + i + "health", sprite);

				sprite = GuiManager.getSingleton.load("player" + i + "kdoutline", "assets/sprites/killdeathoutline");
				sprite.Position = new Vector2((i - 1) * 200 + 128, 26);
				sprite.pause();
				_hudItems.Add("player" + i + "kdoutline", sprite);

				sprite = GuiManager.getSingleton.load("player" + i + "gunavatar", "assets/sprites/gunavatar");
				sprite.Position = new Vector2((i - 1) * 200 + 30, 20);
				sprite.pause();
				_hudItems.Add("player" + i + "gunavatar", sprite);

				msg = new Message("0", new Vector2((i - 1) * 200 + 100, 14), Color.Orange, "ug", "player" + i + "kills");
				GuiManager.getSingleton.load("player" + i + "kills", msg);
				_hudMsgItems.Add("player" + i + "kills", msg);
				msg = new Message("0", new Vector2((i - 1) * 200 + 150, 14), Color.Orange, "ug", "player" + i + "deaths");
				GuiManager.getSingleton.load("player" + i + "deaths", msg);
				_hudMsgItems.Add("player" + i + "deaths", msg);
			}
		}

		/// <summary>
		/// Unloads all the assets for this state. Usually called when shutdown is called.
		/// </summary>
		public override void unload()
		{

		}

		/// <summary>
		/// Tells the engine to render this game state
		/// </summary>
		public override void run()
		{
			GraphicsCore.getSingleton.present();
		}

		/// <summary>
		/// What this state does when it becomes the main state.
		/// </summary>
		public override void enter()
		{
			SceneManager.getSingleton.setActiveScene(GAME_STATE_NAME);

			foreach (Sprite sprite in _hudItems.Values)
			{
				sprite.Visible = true;
			}

			foreach (Message msg in _hudMsgItems.Values)
				msg.visible = true;

			_pause = false;
		}

		/// <summary>
		/// What this state does when it exits.
		/// </summary>
		public override void exit()
		{
			//unload();

			_pause = true;

			foreach (Sprite sprite in _hudItems.Values)
			{
				sprite.Visible = false;
			}

			foreach (Message msg in _hudMsgItems.Values)
				msg.visible = false;
		}

		/// <summary>
		/// Pauses the game state.
		/// </summary>
		/// <returns>
		/// True, if the game is paused. False, if it is running.
		/// </returns>
		public override bool pause()
		{
			_pause = true;

			foreach (Sprite sprite in _hudItems.Values)
			{
				sprite.Visible = false;
			}

			foreach (Message msg in _hudMsgItems.Values)
				msg.visible = false;

			return true;
		}

		/// <summary>
		/// Resumes the game state.
		/// </summary>
		public override void resume()
		{
			_pause = false;

			foreach (Sprite sprite in _hudItems.Values)
			{
				sprite.Visible = true;
			}

			foreach (Message msg in _hudMsgItems.Values)
				msg.visible = true;
		}

		/// <summary>
		/// Updates the game state. Updates parts of the engine that's important to the states. Does state change here.
		/// </summary>
		/// <param name="gameTime">The current values for time.</param>
		public override void update(GameTime gameTime)
		{
			base.update(gameTime);

			playerManager.update(gameTime.ElapsedGameTime.Milliseconds * .001f);
			playerManager.handle();

			// update HUD info
			updateHUDInfo();

			GuiManager.getSingleton.update(gameTime.ElapsedGameTime.Milliseconds * .001f);

			//Rotates the background
			background = SpriteManager.getSingleton.get("background");
			background.Rotation -= 0.0001f;
			if (background.Rotation <= -360)
			{
				background.Rotation += 360;
			}

			foreach (Player player in playerManager.Players.Values)
			{
				if (player.Kills >= 20)
				{
					//Player has won the game!!!
					//_parent.pushGameState(_parent.getGameState(WinGameState.GAME_STATE_NAME));
				}
			}

			//Update the crosshairs position to that of the mousecursor
			mouseCursor.Position = input.current[(int)PlayerIndex.One].mousePosition;

			// check input
			if (Keyboard.GetState().IsKeyDown(Keys.F11))
			{
				GraphicsCore.getSingleton.Graphics.ToggleFullScreen();
			}
			if (Keyboard.GetState().IsKeyDown(Keys.Escape))
			{
				Console.Out.WriteLine(GAME_STATE_NAME + " escape");
				_parent.popGameState();
			}

            ExplosionManager.getSingleton.update(gameTime.ElapsedGameTime.Milliseconds);
		}

		/// <summary>
		/// Updates the HUD info for each player
		/// </summary>
		private void updateHUDInfo()
		{
			foreach (Player player in playerManager.Players.Values)
			{
				int playerNum = (int)player.PlayerNumber + 1;
				Sprite sprite = null;
				Message msg = null;

				if (_hudItems.ContainsKey("player" + playerNum + "health"))
				{
					sprite = _hudItems["player" + playerNum + "health"];
					sprite.setAnimation(player.HealthPoints.ToString());
				}

				if (_hudMsgItems.ContainsKey("player" + playerNum + "kills"))
				{
					msg = _hudMsgItems["player" + playerNum + "kills"];
					msg.text = player.Kills.ToString();
				}

				if (_hudMsgItems.ContainsKey("player" + playerNum + "deaths"))
				{
					msg = _hudMsgItems["player" + playerNum + "deaths"];
					msg.text = player.Deaths.ToString();
				}
			}
		}

		/// <summary>
		/// Ends the program.
		/// </summary>
		public override void shutdown()
		{

		}
	}
}
