/*
 *  Author: David Huynh and Robert Yates
 *  Copyright 2007. All rights reserved.
 */

using System;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;

using rhodes.graphics;
using rhodes.engine;
using rhodes.physics;
using rhodes.audio;
using rhodes.input;

using sedohr.player;

namespace sedohr
{
	/// <summary>
	/// Stores information about the current "screen" being displayed in the
	/// program. For example, a title screen is a game state. If a demo game
	/// state is created and pushed onto the game state manager, the title
	/// screen pauses, the demo game becomes the current game state.
	/// </summary>
    abstract class GameState
    {
		/// <summary>
		/// The manager of this game state.
		/// </summary>
        protected GameStateManager _parent;

		/// <summary>
		/// True, if the game is paused. False, if it is running.
		/// </summary>
        protected Boolean _pause;

		/// <summary>
		/// Creates a game state with a given GameStateManager.
		/// </summary>
		/// <param name="parent">The GameStateManager that is storing this game state.</param>
        public GameState(GameStateManager parent)
        {
            _parent = parent;
        }

		/// <summary>
		/// Loads all the assets for this state.
		/// </summary>
        abstract public void load();

		/// <summary>
		/// Unloads all the assets for this state. Usually called when shutdown is called.
		/// </summary>
        abstract public void unload();
 
		/// <summary>
		/// Updates the game state. Updates parts of the engine that's important to the states. Does state change here.
		/// </summary>
		/// <param name="gameTime">The current values for time.</param>
		virtual public void update(GameTime gameTime)
		{
			/* Update rhodes engine */
			PhysicsManager.getSingleton.update(gameTime.ElapsedGameTime.Milliseconds * .001f);
			EntityManager.getSingleton.update(gameTime.ElapsedGameTime.Milliseconds * .001f);
			Camera2D.getSingleton.update(gameTime.ElapsedGameTime.Milliseconds * .001f);
			AudioManager.getSingleton.update();
			SceneManager.getSingleton.Update();
            ParticleManager.getSingleton.update(gameTime.ElapsedGameTime.Milliseconds * .001f);
		}

		/// <summary>
		/// Tells the engine to render this game state
		/// </summary>
        abstract public void run();

		/// <summary>
		/// What this state does when it becomes the main state.
		/// </summary>
        abstract public void enter();

		/// <summary>
		/// What this state does when it exits.
		/// </summary>
        abstract public void exit();

		/// <summary>
		/// Pauses the game state.
		/// </summary>
		/// <returns>
		/// True, if the game is paused. False, if it is running.
		/// </returns>
        abstract public bool pause();

		/// <summary>
		/// Resumes the game state.
		/// </summary>
        abstract public void resume();

		/// <summary>
		/// Returns the game state of a specific name.
		/// </summary>
		/// <param name="name">The name of the game state.</param>
		/// <returns>The game state itself.</returns>
        protected GameState getGameState(String name)
        {
            return _parent.getGameState(name);
        }

		/// <summary>
		/// Returns the game state of a specific name.
		/// </summary>
		/// <param name="name">The name of the game state.</param>
		/// <returns>The game state itself.</returns>
        protected void changeGameState(GameState state)
        {
            _parent.changeGameState(state);
        }
        
		/// <summary>
		/// Adds a game state into the game state manager of
		/// this game state. The current game state is paused
		/// while the new game state is initialized.
		/// </summary>
		/// <param name="state">The game state to add.</param>
		/// <returns>
		/// True, if this game state was added. False, it it was not added.
		/// </returns>
        public bool pushGameState(GameState state) 
        {
            return _parent.pushGameState(state);
        }

		/// <summary>
		/// Removes the current game state from the game state manager of
		/// this game state. The previous game state on the stack resumes.
		/// </summary>
        public void popGameState() 
        { 
            _parent.popGameState(); 
        }
		
		/// <summary>
		/// Retreives an instance of the input.
		/// </summary>
		/// <returns>The input.</returns>
		public Input getInput()
		{
			return this._parent.getInput();
		}
        
		/// <summary>
		/// Ends the program.
		/// </summary>
        abstract public void shutdown();
    }
}
