/*
 *  Author: David Huynh and Robert Yates
 *  Copyright 2007. All rights reserved.
 */

using System;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;

using rhodes.input;

namespace sedohr
{

	/// <summary>
	/// Manages multiple game states.
	/// </summary>
    class GameStateManager
    {
		/// <summary>
		/// All the game states.
		/// </summary>
        private Dictionary<String, GameState> _states;
		/// <summary>
		/// The stack of all the states.
		/// The running state is at the top, while the
		/// remaining states are paused.
		/// </summary>
        private Stack<GameState> _stateStack;

		/// <summary>
		/// The input to control the game.
		/// </summary>
		private Input _input;

		/// <summary>
		/// The player who controls the menu screens.
		/// (Defaults to player one.)
		/// </summary>
		private PlayerIndex playerControllingMenu = PlayerIndex.One;
		/// <summary>
		/// The player who controls the menu screens.
		/// </summary>
		public PlayerIndex PlayerControllingMenu
		{
			get { return playerControllingMenu; }
		}

		/// <summary>
		/// True, if the game is shutting down. False, if not.
		/// </summary>
		private Boolean _shutdown;

		/// <summary>
		/// True, if the game is shutting down. False, if not.
		/// </summary>
		public Boolean Shutdown {
			get { return _shutdown; }
		}

		/// <summary>
		/// Creates a game state manager.
		/// </summary>
        public GameStateManager()
        {
            _states = new Dictionary<string, GameState>();
            _stateStack = new Stack<GameState>();
            _shutdown = false;
			_input = new Input();

			// Make it so the lowest number player that is connected
			// controllers the menu
			if (_input.current[(int)PlayerIndex.One].connected)
			{
				playerControllingMenu = PlayerIndex.One;
			}
			else if (_input.current[(int)PlayerIndex.Two].connected)
			{
				playerControllingMenu = PlayerIndex.Two;
			}
			else if (_input.current[(int)PlayerIndex.Three].connected)
			{
				playerControllingMenu = PlayerIndex.Three;
			}
			else if (_input.current[(int)PlayerIndex.Four].connected)
			{
				playerControllingMenu = PlayerIndex.Four;
			}
			else
			{
				// default to player 1 (keyboard) if
				// no Xbox 360 controllers are plagged in
				playerControllingMenu = PlayerIndex.One;
			}

        }

		/// <summary>
		/// Registers a game state with the game state manager.
		/// </summary>
		/// <param name="name">The name of the game state.</param>
		/// <param name="state">The state itself.</param>
        public void addGameState(String name, GameState state)
        {
            if (!_states.ContainsKey(name))
            {
                _states.Add(name, state);
            }
        }

		/// <summary>
		/// Removed a game state from the game state maanger.
		/// </summary>
		/// <param name="name">The name of the game state.</param>
		public void removeGameState(String name)
		{
			if (_states.ContainsKey(name))
			{
				_states.Remove(name);
			}
		}

		/// <summary>
		/// Returns the game state of a specific name.
		/// </summary>
		/// <param name="name">The name of the game state.</param>
		/// <returns>The game state.</returns>
        public GameState getGameState(String name)
        {
            if (_states.ContainsKey(name))
                return _states[name];

            return null;
        }

		/// <summary>
		/// Updates the input and the current game state.
		/// </summary>
		/// <param name="time">The current values for time.</param>
		/// <returns>
		/// True, if the game state was updated. False, if not.
		/// </returns>
        public bool update(GameTime time)
        {
            if (_shutdown)
                return false;

			_input.update();

			_stateStack.Peek().update(time);

            return true;
        }

		/// <summary>
		/// Runs the current game state.
		/// </summary>
        public void run()
        {
			if (_shutdown)
				return;

            _stateStack.Peek().run();
        }

		/// <summary>
		/// Exits the current game state and replaces it with the
		///  new game state. The new game state is started.
		/// </summary>
		/// <param name="state">The new game state.</param>
        public void changeGameState(GameState state)
        {
            if (_stateStack.Count != 0)
            {
                _stateStack.Peek().exit();
                _stateStack.Pop();
            }

            _stateStack.Push(state);
            _stateStack.Peek().enter();
        }

		/// <summary>
		/// Pushes the current game state onto the game state stack.
		/// </summary>
		/// <param name="state">The game state to add.</param>
		/// <returns>
		/// True, if the game state was added. False if it wasn't.
		/// </returns>
        public bool pushGameState(GameState state)
        {
            // pause current game state
            if (_stateStack.Count != 0)
            {
				_stateStack.Peek().pause();
            }

            // store and init the new state
            _stateStack.Push(state);
            _stateStack.Peek().enter();

            return true;
        }

		/// <summary>
		/// Removes the current game state from the game state manager of
		/// this game state. The previous game state on the stack resumes.
		/// </summary>
        public void popGameState()
        {
            // pop the state
            if (_stateStack.Count != 0)
            {
                _stateStack.Peek().exit();
                _stateStack.Pop();
            }

            // resume previous game state
            if (_stateStack.Count != 0)
            {
                _stateStack.Peek().resume();
            }
            else
            {
                shutdown();
            }
        }

		/// <summary>
		/// Retreives an instance of the input.
		/// </summary>
		/// <returns>The input.</returns>
		public Input getInput()
		{
			return _input;
		}

		/// <summary>
		/// Ends the program.
		/// </summary>
        public void shutdown()
        {
            _shutdown = true;
        }
    }
}
