using System;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;

using FarseerGames.FarseerPhysics;
using FarseerGames.FarseerPhysics.Dynamics;
using FarseerGames.FarseerPhysics.Collisions;

using RhodesProcessorLib;

using rhodes.graphics;
using rhodes.engine;
using rhodes.physics;
using rhodes.audio;
using rhodes.input;

namespace sedohr
{
	/// <summary>
	/// This class initializes everything related to the game
	/// (ex. graphics, camera, physics, audio) and the title screeen is displayed
	/// Resource used: http://www.nuclex.org/articles/using-xna-to-draw-a-rotating-triangle
	/// Author: Robert Yates
    /// Modified: Josh Wilson
	/// </summary>
	class SedohrGame : Microsoft.Xna.Framework.Game
	{
		/// <summary>
		/// True, if the game should start in fullscreen mode. False, if not.
		/// </summary>
		private const bool START_IN_FULLSCREEN = false;

		/// <summary>
		/// The screen's horizontal resolution
		/// </summary>
		private const int HORIZONTAL_RESOLUTION = 800;

		/// <summary>
		/// The screen's vertical resolution
		/// </summary>
		private const int VERTICAL_RESOLUTION = 600;

		/// <summary>
		/// Draws objects to the screen.
		/// </summary>
		GraphicsDeviceManager graphics;
		/// <summary>
		/// Loads content from binary files in content pipeline.
		/// </summary>
		ContentManager content;

		/// <summary>
		/// Manages multiples game states (ex. TitleGameState, DemoGameState)
		/// </summary>
		GameStateManager gameStateManager;

		/// <summary>
		/// Creates graphics, content and gameStateManager.
		/// </summary>
		public SedohrGame()
		{
			graphics = new GraphicsDeviceManager(this);
			content = new ContentManager(Services);
			gameStateManager = new GameStateManager();
		}

		/// <summary>
		/// Allows the game to perform any initialization it needs to before
		/// starting to run. This method enumerates through any components
		/// and initializes them as well.
		/// </summary>
		protected override void Initialize()
		{
			graphics.PreferredBackBufferWidth = HORIZONTAL_RESOLUTION;
			graphics.PreferredBackBufferHeight = VERTICAL_RESOLUTION;
			graphics.MinimumPixelShaderProfile = ShaderProfile.PS_2_0;
			graphics.IsFullScreen = START_IN_FULLSCREEN;
			graphics.ApplyChanges();

			GraphicsCore.getSingleton.initialize(graphics, content);
			Camera2D.getSingleton.initialize(graphics.PreferredBackBufferWidth,
				graphics.PreferredBackBufferHeight);
			PhysicsManager.getSingleton.initialize(new Vector2(0.0f, 700.0f));
			AudioManager.getSingleton.initialize();

			base.Initialize();
		}

		/// <summary>
		/// Load graphics content from resource management pools.
		/// </summary>
		protected override void LoadContent()
		{
			GameState titleGameState = new TitleGameState(gameStateManager);
			titleGameState.load();
			gameStateManager.addGameState(TitleGameState.GAME_STATE_NAME, titleGameState);

			GameState helpGameState = new HelpGameState(gameStateManager);
			helpGameState.load();
			gameStateManager.addGameState(HelpGameState.GAME_STATE_NAME, helpGameState);

			GameState pauseGameState = new PauseGameState(gameStateManager);
			pauseGameState.load();
			gameStateManager.addGameState(PauseGameState.GAME_STATE_NAME, pauseGameState);

			GameState creditGameState = new CreditGameState(gameStateManager);
			creditGameState.load();
			gameStateManager.addGameState(CreditGameState.GAME_STATE_NAME, creditGameState);

			GameState winGameState = new WinGameState(gameStateManager);
			winGameState.load();
			gameStateManager.addGameState(WinGameState.GAME_STATE_NAME, winGameState);

			gameStateManager.pushGameState(titleGameState);
		}

		/// <summary>
		/// Unload graphics content from resource management pools.
		/// </summary>
		protected override void UnloadContent()
		{
			GraphicsCore.getSingleton.Content.Unload();
		}


		/// <summary>
		/// Allows the game to run logic such as updating the world,
		/// checking for collisions, gathering input and playing audio.
		/// </summary>
		/// <param name="gameTime">Contains information about the elapsed time.</param>
		protected override void Update(GameTime gameTime)
		{
			gameStateManager.update(gameTime);
			base.Update(gameTime);
		}

		/// <summary>
		/// Render the menu or game world
		/// </summary>
		/// <param name="gameTime">Contains information about the elapsed time</param>
		protected override void Draw(GameTime gameTime)
		{
			if (gameStateManager.Shutdown)
				this.Exit();

			gameStateManager.run();

			base.Draw(gameTime);
		}
	}
}
