/*
 *  Author: Josh Wilson
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 */

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;

using rhodes.graphics;
using rhodes.input;

using sedohr.player;

namespace sedohr
{
	/// <summary>
	/// This is the state that the game will enter when a player wins the game. It shows a close up of the player who won the game.s
	/// </summary>
    class WinGameState : GameState
    {
        /// <summary>
        /// Contains the name of the game state
        /// </summary>
        public readonly static string GAME_STATE_NAME = "win";

        /// <summary>
        /// Contains a list of all message items in the title
        /// </summary>
        private SortedDictionary<String, Message> _winMsgItems;

        /// <summary>
        /// Contains a list of the menu options
        /// </summary>
        private List<Message> _menu;

        /// <summary>
        /// Menu option that is currently selected
        /// </summary>
        private int _menuSelection = 0;

        /// <summary>
        /// Contains a list of all message items in the title
        /// </summary>
        private SortedDictionary<String, Sprite> _winItems;

		/// <summary>
		/// The input component of this state
		/// </summary>
		private Input _input;

		/// <summary>
		/// The player who won the game.
		/// </summary>
		private Player _winningPlayer;

		/// <summary>
		/// Text for a message.
		/// </summary>
		private string _text;

		/// <summary>
		/// Color of message.
		/// </summary>
		private Vector4 _color;

        /// <summary>
        /// Win Game State Constructor
        /// </summary>
        /// <param name="parent">Instance of the GameStateManager</param>
        public WinGameState(GameStateManager parent)
            : base(parent)
        {
            _winMsgItems = new SortedDictionary<string, Message>();
            _winItems = new SortedDictionary<string, Sprite>();
            _menu = new List<Message>();
			_input = parent.getInput();
        }

        /// <summary>
        /// Loads the Win Game State
        /// </summary>
        public override void load()
        {
			//Create the graphics for the Win Game State
			Sprite sprite = GuiManager.getSingleton.load("winbox", "assets/sprites/pause");
			sprite.pause();
			sprite.Position = new Vector2(400, 300);
			_winItems.Add("winbox", sprite);

			 sprite = GuiManager.getSingleton.load("winimage", "assets/sprites/winning_screen");
			sprite.pause();
			sprite.Position = new Vector2(400, 300);
			_winItems.Add("winimage", sprite);

			
            Message msg = new Message(_text, new Vector2(50, 25), Color.White, "ug", "wininfo");
			msg.scale = 2;
			GuiManager.getSingleton.load("wininfo", msg);
            _winMsgItems.Add("wininfo", msg);

            msg = new Message("Exit", new Vector2(200, 450), Color.Red, "ug", "winexit");
            GuiManager.getSingleton.load("winexit", msg);
            _menu.Add(msg);

            pause();
        }

        /// <summary>
        /// Unloads the Win Game State
        /// </summary>
        public override void unload()
        {

        }

        /// <summary>
        /// Renders the Win Game State
        /// </summary>
        public override void run()
        {
			GraphicsCore.getSingleton.present();
        }

        /// <summary>
        /// Enters the Win Game State
        /// </summary>
        public override void enter()
        {
            foreach (Sprite sprite in _winItems.Values)
            {
                sprite.Visible = true;
            }

            foreach (Message msg in _menu)
                msg.visible = true;

            foreach (Message msg in _winMsgItems.Values)
                msg.visible = true;

			GuiManager.getSingleton.getMessage("wininfo").text = _text;

            _pause = false;
        }

        /// <summary>
        /// Exits the Win Game State
        /// </summary>
        public override void exit()
        {
            foreach (Sprite sprite in _winItems.Values)
            {
                sprite.Visible = false;
            }

            foreach (Message msg in _menu)
                msg.visible = false;

            foreach (Message msg in _winMsgItems.Values)
                msg.visible = false;
        }

        /// <summary>
        /// Pauses the Win Game State
        /// </summary>
        /// <returns></returns>
        public override bool pause()
        {
            _pause = true;

            foreach (Sprite sprite in _winItems.Values)
            {
                sprite.Visible = false;
            }

            foreach (Message msg in _menu)
                msg.visible = false;

            foreach (Message msg in _winMsgItems.Values)
                msg.visible = false;

            return true;
        }

        /// <summary>
        /// Resumes the Win Game State
        /// </summary>
        public override void resume()
        {
            foreach (Sprite sprite in _winItems.Values)
            {
                sprite.Visible = true;
            }

            foreach (Message msg in _menu)
                msg.visible = true;

            foreach (Message msg in _winMsgItems.Values)
                msg.visible = true;

            _pause = false;
        }

        /// <summary>
        /// Updates the Win Game State
        /// </summary>
        /// <param name="gameTime">Snapshot of the game timing state</param>
        public override void update(GameTime gameTime)
        {
            if (_pause)
                return;

            GuiManager.getSingleton.update(gameTime.ElapsedGameTime.Milliseconds * .001f);

            //When the player presses up, change to the appropriate menu selection
			if (_input.current[(int)_parent.PlayerControllingMenu].up == ButtonState.Pressed &&
				_input.previous[(int)_parent.PlayerControllingMenu].up == ButtonState.Released)
			{
				_menu[_menuSelection].color = Color.Yellow;

				_menuSelection--;

				if (_menuSelection < 0)
					_menuSelection = _menu.Count - 1;

				_menu[_menuSelection].color = Color.Red;
			}

            //When the player presses down, change to the appropriate menu selection
			if (_input.current[(int)_parent.PlayerControllingMenu].down == ButtonState.Pressed &&
				_input.previous[(int)_parent.PlayerControllingMenu].down == ButtonState.Released)
			{
				_menu[_menuSelection].color = Color.Yellow;

				_menuSelection++;

				if (_menuSelection >= _menu.Count)
					_menuSelection = 0;

				_menu[_menuSelection].color = Color.Red;
			}

            //When the player presses enter(start on the controller), change to the appropriate game state
			if (_input.current[(int)_parent.PlayerControllingMenu].pause == ButtonState.Pressed &&
				_input.previous[(int)_parent.PlayerControllingMenu].pause == ButtonState.Released)
			{
                switch (_menuSelection)
                {
                    case 0:
                        //Go to back to Main Menu
                        foreach (Message msg in _menu)
                            msg.visible = false;
                        foreach (Message msg in _winMsgItems.Values)
                            msg.visible = false;
                        _parent.popGameState();
						_parent.popGameState();
						_parent.popGameState();
                        break;
                }
            }
        }

        /// <summary>
        /// Shuts down the Win Game State
        /// </summary>
        public override void shutdown()
        {

        }

		/// <summary>
		/// Sets the winning info for the player who won the game
		/// </summary>
		/// <param name="player"></param>
		public void setWinningPlayerInfo(Player player)
		{
			_winningPlayer = player;
			_text = "CONGRATULATIONS Player " + _winningPlayer.PlayerNumber + "!!!";
			_winItems["winimage"].Color = player.Entity.Sprite.Color;
			GuiManager.getSingleton.getMessage("wininfo").color = new Color(player.Entity.Sprite.Color);
		}
    }
}