/*
 *  Author: Josh Wilson
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 */

using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using FarseerGames.FarseerPhysics;
using FarseerGames.FarseerPhysics.Dynamics;
using FarseerGames.FarseerPhysics.Collisions;

using rhodes.graphics;
using rhodes.physics;
using rhodes.engine;

using sedohr.player;
using sedohr.weapon;

namespace sedohr.objects
{
    /// <summary>
    /// Class used to implement the Black Hole.
	/// NOTE*** This Class is not being used in the current implementation of the game. This class will possibly be used in
	///		future implementations of the game.
    /// </summary>
	class Hole
	{
        /// <summary>
        /// Duration the black hole is on the screen
        /// </summary>
		private const float ACTIVE_DURATION = 5.0f;

        /// <summary>
        /// Duration the black hole is off the screen
        /// </summary>
		private const float INACTIVE_DURATION = 30.0f;

        /// <summary>
        /// The strength of the black hole
        /// </summary>
        private const float STRENGTH = 50000000.0f; //250000.0f; 

		/// <summary>
		/// A point off the screen where the hole despawned
		/// </summary>
		private Vector2 DESPAWN_POINT = new Vector2(-6000.0f, -6000.0f);

        /// <summary>
        /// Name of the Black/White hole.
        /// </summary>
		private String _name;
		public String Name
		{
			get { return _name; }
			set { _name = value; }
		}

        /// <summary>
        /// Whether the Black hole is active.
        /// </summary>
		private Boolean _active;

        /// <summary>
        /// List of all the players sucked into the Black hole.
        /// </summary>
		private SortedDictionary<string, Player> _capturedPlayers;
		public SortedDictionary<string, Player> CapturedPlayers
		{
			get { return _capturedPlayers; }
		}

        /// <summary>
        /// List of all the simple objects sucked into the black hole.
        /// </summary>
		private SortedDictionary<string, SimpleObject> _capturedSimpleObjects;
		public SortedDictionary<string, SimpleObject> CapturedSimpleObjects
		{
			get { return _capturedSimpleObjects; }
		}

        /// <summary>
        /// List of all the compound objects sucked into the black hole.
        /// </summary>
		private SortedDictionary<string, CompoundObject> _capturedCompoundObjects;
		public SortedDictionary<string, CompoundObject> CapturedCompoundOjbects
		{
			get { return _capturedCompoundObjects; }
		}

        /// <summary>
        /// The black hole entity.
        /// </summary>
		private SimpleEntity _entity;
		public SimpleEntity Entity
		{
			get { return _entity; }
			set { _entity = value; }
		}

        /// <summary>
        /// Duration the black hole is active.
        /// </summary>
		private float _activeDuration;
		public float ActiveDuration
		{
			get { return _activeDuration; }
			set { _activeDuration = value; }
		}

        /// <summary>
        /// The duration the black hole is inactive.
        /// </summary>
		private float _inactiveDuration;
		public float InactiveDuration
		{
			get { return _inactiveDuration; }
			set { _inactiveDuration = value; }
		}

        /// <summary>
        /// List of all the possible spawn points of the black/white hole.
        /// </summary>
		private List<Vector2> _spawnPoints;

        /// <summary>
        /// An instance of the player manager to cycle through the players upon updating.
        /// </summary>
        private PlayerManager _playerManager;

        /// <summary>
        /// An instance of the object manager to cycle through the objects upon updating.
        /// </summary>
        private ObjectManager _objectManager;

        /// <summary>
        /// Hole Constructor.
        /// </summary>
        /// <param name="name">Name of the Black/White Hole.</param>
        /// <param name="playerManager">Instance of the player manager</param>
        public Hole(String name, PlayerManager playerManager, ObjectManager objectManager)
        {
			_name = name;
            _playerManager = playerManager;
            _objectManager = objectManager;
			_inactiveDuration = INACTIVE_DURATION;
			_activeDuration = ACTIVE_DURATION;

			_capturedCompoundObjects = new SortedDictionary<string, CompoundObject>();
			_capturedPlayers = new SortedDictionary<string, Player>();
			_capturedSimpleObjects = new SortedDictionary<string, SimpleObject>();
			_spawnPoints = new List<Vector2>();

            //Load the black hole entity
			_entity = EntityManager.getSingleton.loadSimple("assets/entities/hole");

            //Create the Collisions geometry for the black hole
            foreach (Geom geom in _entity.PhysicsBody.GeomList)
            {
                geom.CollisionHandler += this.collisionCallback;
            }
			
            //Create the spawn points for the black hole
			_spawnPoints.Add(new Vector2(860.0f, 360.0f));
			_spawnPoints.Add(new Vector2(1515.0f, 464.0f));
			_spawnPoints.Add(new Vector2(2204.0f, 308.0f));
			_spawnPoints.Add(new Vector2(2492.0f, 904.0f));
			_spawnPoints.Add(new Vector2(428.0f, 892.0f));
			_spawnPoints.Add(new Vector2(1008.0f, 824.0f));
			_spawnPoints.Add(new Vector2(1900.0f, 840.0f));
			_spawnPoints.Add(new Vector2(2160.0f, 1320.0f));
			_spawnPoints.Add(new Vector2(1440.0f, 1276.0f));
			_spawnPoints.Add(new Vector2(908.0f, 1364.0f));
			_spawnPoints.Add(new Vector2(1436.0f, 1676.0f));

            //Set the black hole to initially be inactive (off the screen)
			Active = false;
        }

        /// <summary>
        /// Updates the hole.
        /// </summary>
        /// <param name="elapsedTime">Elapsed time in seconds since the last update.</param>
        public void update(float elapsedTime)
        {
            //If the black hole is active (on the screen)
            if (_active)
            {
                //Update it's active duration and despawn it if the duration is up
                _activeDuration -= elapsedTime;
                if (_activeDuration < 0)
                {
                    Despawn();
                }
				Vector2 dir;
				Vector2 force;
				float distance;
				float invprop;

				#region affect players (commented out)
				//NOTE*** Don't affect the players
                //Cycle through the players applying a force towards the black hole
				//foreach (Player player in _playerManager.Players.Values)
				//{
				//    //if (STRENGTH != 0 && !player.Entity.PhysicsBody.Body.IgnoreGravity)
				//    //    player.Entity.PhysicsBody.Body.IgnoreGravity = true;
				//    if(player.Entity.Active)
				//    {
				//        dir = _entity.PhysicsBody.Body.Position - player.Entity.PhysicsBody.Body.Position;
				//        distance = dir.Length(); //(float)Math.Sqrt((Math.Pow(dir.X, 2) + Math.Pow(dir.Y, 2)));
				//        invprop = (float)(STRENGTH / (Math.Pow(distance, 2) * (Math.Pow(player.Entity.PhysicsBody.Body.Mass, 2))));
				//        //float invprop = (float)(player.Entity.PhysicsBody.Body.Mass * STRENGTH / Math.Pow(distance, 2));
				//        force = Vector2.Multiply(dir, invprop); //(player.Entity.PhysicsBody.Body.Mass / Math.Pow(distance, 2)));

				//        player.Entity.PhysicsBody.Body.ApplyForce(force);

				//        //foreach (LaserBeam laserbeam in player.Weapons["laser"].ProjectilesOrLaserBeams.Values)
				//        //{
				//        //    if (laserbeam.Active)
				//        //    {
				//        //        dir = _entity.PhysicsBody.Body.Position - laserbeam.PhysicsBody.Body.Position;
				//        //        distance = dir.Length();
				//        //        invprop = (float)((STRENGTH * 100000000) / (Math.Pow(distance, 2) * player.Entity.PhysicsBody.Body.Mass));
				//        //        force = Vector2.Multiply(dir, invprop);

				//        //        laserbeam.PhysicsBody.Body.ApplyForce(force);
				//        //    }
				//        //}
				//    }
				//}
				#endregion

				//Loop through each rocket projectile and if its active apply a force to it
				foreach (Projectile projectile in ProjectileManager.getSingleton.RocketProjectiles.Values)
				{
					if (projectile.Active)
					{
						dir = _entity.PhysicsBody.Body.Position - projectile.PhysicsBody.Body.Position;
						distance = dir.Length();
						invprop = (float)(STRENGTH / (Math.Pow(distance, 2) * projectile.PhysicsBody.Body.Mass));
						force = Vector2.Multiply(dir, invprop);
						projectile.PhysicsBody.Body.ApplyForce(force);
					}
				}

                //Loop through each pistol projectile and if its active apply a force to it
				foreach (Projectile projectile in ProjectileManager.getSingleton.PistolProjectiles.Values)
				{
					if (projectile.Active)
					{
						dir = _entity.PhysicsBody.Body.Position - projectile.PhysicsBody.Body.Position;
						distance = dir.Length(); 
						invprop = (float)(STRENGTH / (Math.Pow(distance, 2) * projectile.PhysicsBody.Body.Mass));
						force = Vector2.Multiply(dir, invprop); 
						projectile.PhysicsBody.Body.ApplyForce(force);
					}
				}

				#region affect objects (commented out)
				//Loop through each simple object and if its active apply a force to it
				//foreach (SimpleObject simpleObject in _objectManager.Simples.Values)
				//{
				//    //Apply force to each object
				//    if(simpleObject.Active)
				//    {
				//        dir = _entity.PhysicsBody.Body.Position - simpleObject.SimpleEntity.PhysicsBody.Body.Position;
				//        distance = dir.Length();
				//        invprop = (float)(STRENGTH / (Math.Pow(distance, 2) * simpleObject.SimpleEntity.PhysicsBody.Body.Mass));
				//        force = Vector2.Multiply(dir, invprop);
				//        simpleObject.SimpleEntity.PhysicsBody.Body.ApplyForce(force);
				//    }
				//}

                //Note* Do not affect Comnpound objects
				//foreach(CompoundObject compoundObject in _objectManager.Compounds.Values)
				//{
				//    //Apply force to each simple object inside the compound object
				//    if (compoundObject.Active)
				//    {
				//        //dir = _entity.PhysicsBody.Body.Position - projectile.PhysicsBody.Body.Position;
				//        //distance = dir.Length(); //(float)Math.Sqrt((Math.Pow(dir.X, 2) + Math.Pow(dir.Y, 2)));
				//        //invprop = (float)(STRENGTH / (Math.Pow(distance, 2) * projectile.PhysicsBody.Body.Mass));
				//        ////float invprop = (float)(projectile.PhysicsBody.Body.Mass * STRENGTH / Math.Pow(distance, 2));
				//        //force = Vector2.Multiply(dir, invprop); //(player.Entity.PhysicsBody.Body.Mass / Math.Pow(distance, 2)));

				//        //projectile.PhysicsBody.Body.ApplyForce(force);
				//    }
				//}
				#endregion
			}
            else
            {
                //Update it's inactive duration and respawn it if the duration is up
                _inactiveDuration -= elapsedTime;
                if (_inactiveDuration < 0)
                {
					//Console.Out.WriteLine("HOLE APPEARED!!!");
                    Respawn();
				}

				#region affect players (commented out)
				//Don't affect the players
                //Cycle through all the players and let them be affect by gravity again
				//foreach (Player player in _playerManager.Players.Values)
				//{
				//    //if (player.Entity.PhysicsBody.Body.IgnoreGravity)
				//    //    player.Entity.PhysicsBody.Body.IgnoreGravity = false;
				//}
				#endregion
			}
        }

        /// <summary>
        /// Handles the collision when something collides with the black hole
        /// </summary>
        /// <param name="g1">The hole geometry</param>
        /// <param name="g2">The geometry of the other object</param>
        /// <param name="contactList">The list of contact points</param>
        /// <returns></returns>
        public bool collisionCallback(Geom g1, Geom g2, ContactList contactList)
        {
            string g2name = g2.Tag.ToString();

            //If the object that collided with the black hole is a player then kill that player
            if (g2name.Equals("player1") || g2name.Equals("player2") || g2name.Equals("player3") || g2name.Equals("player4"))
            {
                _playerManager.Players[g2name].takeDamage(101);
            }
            else if (g2.CollisionCategories == Enums.CollisionCategories.Cat6)
            {
				
            }
            return true;
        }

        /// <summary>
        /// Gets and Sets the active property
        /// </summary>
		public Boolean Active
		{
			get
			{
				return _active;
			}
			set
			{
				if (_active != value)
				{
					_active = value;

					if (_active)
					{
						_entity.Active = true;
					}
					else
					{
						_entity.Active = false;
					}
				}
			}
		}

        /// <summary>
        /// Add a Player to the list of players sucked into the black hole
        /// </summary>
        /// <param name="player">Player sucked into the black hole</param>
		public void AddPlayer(Player player)
		{
			if (_capturedPlayers.ContainsKey(player.PlayerName))
			{
				Console.Error.WriteLine("Player " + player.PlayerName + " already exists in hole.");
			}
			else
			{
				_capturedPlayers.Add(player.PlayerName, player);
			}
		}

        /// <summary>
        /// Add a simple object to the list of simple objects sucked into the black hole
        /// </summary>
        /// <param name="simpleObject">The simple object sucked into the black hole</param>
		public void AddSimpleObject(SimpleObject simpleObject)
		{
			if(_capturedSimpleObjects.ContainsKey(simpleObject.Name))
			{
				Console.Error.WriteLine("SimpleObject " + simpleObject.Name + " already exists in hole.");
			}
			else
			{
				_capturedSimpleObjects.Add(simpleObject.Name, simpleObject);
			}
		}

        /// <summary>
        /// Add a compound object to the list of compound objects sucked into the black hole
        /// </summary>
        /// <param name="compoundObject">The compound object sucked into the black hole</param>
		public void AddCompoundObject(CompoundObject compoundObject)
		{
			if(_capturedCompoundObjects.ContainsKey(compoundObject.Name))
			{
				Console.Error.WriteLine("CompoundObject " + compoundObject.Name + " already exists in hole.");
			}
			else
			{
				_capturedCompoundObjects.Add(compoundObject.Name, compoundObject);
			}
		}

        /// <summary>
        /// Gets a player from the list of players in the black hole
        /// </summary>
        /// <param name="name">The name of the player to get</param>
        /// <returns>The player if it is in the list, Null if it isn't</returns>
		public Player GetPlayer(string name)
		{
			if (_capturedPlayers.ContainsKey(name))
			{
				return _capturedPlayers[name];
			}

			return null;
		}

        /// <summary>
        /// Gets a simple object from the list of simple objects in the black hole
        /// </summary>
        /// <param name="name">The name of the simple object</param>
        /// <returns>The simple object if it is in the list, Null if it isn't</returns>
		public SimpleObject GetSimpleObject(string name)
		{
			if(_capturedSimpleObjects.ContainsKey(name))
			{
				return _capturedSimpleObjects[name];
			}

			return null;
		}

        /// <summary>
        /// Gets a compound object from the list of compound objects in the black hole
        /// </summary>
        /// <param name="name">The name of the compound object</param>
        /// <returns>The compound object if it is in the list, Null if it isn't</returns>
		public CompoundObject GetCompoundObject(string name)
		{
			if (_capturedCompoundObjects.ContainsKey(name))
			{
				return _capturedCompoundObjects[name];
			}

			return null;
		}

        /// <summary>
        /// Respawns the black hole back onto the screen and makes it active
        /// </summary>
		public void Respawn()
		{
            //Picks a spawn point at random and spawns the black hole at that point
			Random rand = new Random();
			Vector2 spawnPoint = _spawnPoints[rand.Next(0, 8)];

			_entity.PhysicsBody.Body.Position = spawnPoint;
			_entity.Sprite.Position = spawnPoint;
			_inactiveDuration = INACTIVE_DURATION;
			Active = true;
		}

        /// <summary>
        /// Despawns the black hole and moves it off the screen
        /// </summary>
		public void Despawn()
		{
			_entity.PhysicsBody.Body.Position = DESPAWN_POINT;
			_entity.Sprite.Position = DESPAWN_POINT;
			_activeDuration = ACTIVE_DURATION;
			Active = false;
		}
	}
}
