/*
 *  Author: Josh Wilson, Anthony Reese
 *  Copyright 2007. All rights reserved.
 */

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using FarseerGames.FarseerPhysics;
using FarseerGames.FarseerPhysics.Collisions;
using rhodes.engine;
using rhodes.graphics;
using rhodes.graphics.particles;
using rhodes.graphics.particles.emitters;
using rhodes.input;
using rhodes.physics;
using sedohr.weapon;

namespace sedohr.player
{
	/// <summary>
	/// Player State for when they are dead
	/// </summary>
	class DeadPlayerState : PlayerState
	{
		#region Constants

		/// <summary>
        /// Duration of time a player will stay dead before they respawn
        /// </summary>
		private const float DEATHWAIT = 5.0f;

		#endregion

		#region Attributes

		/// <summary>
        /// Current amount of time the player has been dead
        /// </summary>
		private float _time;

		/// <summary>
		/// Particle effect shown when a player spawns.
		/// </summary>
		private ParticleEngine _spawnEffect;

		/// <summary>
		/// Layer for showing particle effects
		/// </summary>
		private static Layer _particleLayer;

		#endregion

		#region Constructor

		/// <summary>
        /// Dead Player State Constructor
        /// </summary>
        /// <param name="player">Instance of the Player</param>
		public DeadPlayerState(Player player)
			: base(player)
		{
			Name = "Dead";
			_time = 0;

			if (_particleLayer == null)
			{
				_particleLayer = SceneManager.getSingleton.getActiveScene().getLayer(5);
			}
			_spawnEffect = ParticleManager.getSingleton.load("spawn", "assets/particles/spawnEffect");
			_particleLayer.AddParticle("spawn" + player.Entity.Name, _spawnEffect);
			_spawnEffect.register(player.Entity.Sprite.Name);
		}

		#endregion

		#region Functions

		/// <summary>
        /// Updates the Dead Player State
        /// </summary>
        /// <param name="elapsedTime">The amount of time that has passed</param>
		public override void update(float elapsedTime)
		{
            //Increments the total time the player has been dead and if that time exceeds the duration
            // a player stays dead then respawn the player
			_time += elapsedTime;
			
			if (_time > DEATHWAIT)
			{
				_player.changeState("OnGround");
			}
		}

        /// <summary>
        /// Handles Everything
        /// </summary>
		public override void handle() { }

        /// <summary>
        /// Enters the Dead Player State
        /// </summary>
		public override void enter()
		{
			// Move the player off the screen and unregister the player from the Camera
			Vector2 deadPos = new Vector2(-10000.0f, -10000.0f);

			Console.Out.WriteLine(_player.PlayerName + " DEAD");

			_player.Entity.PhysicsBody.Body.ClearForce();
			_player.Entity.PhysicsBody.Body.LinearVelocity = Vector2.Zero;
			_player.Entity.PhysicsBody.Body.Position = deadPos;
			_player.Entity.Sprite.Position = deadPos;
			_player.CurrentWeapon.Sprite.Position = deadPos;
			Camera2D.getSingleton.unregister(_player.Entity.Name);
			_player.Entity.Active = false;
			_player.HealthPoints = 0;
			foreach (Weapon weapon in _player.Weapons.Values)
			{
				if (weapon.Genre == Weapon.weaponType.LASER)
				{
					((Laser)weapon).destroyLaserBeam();
				}
			}
		}

        /// <summary>
        /// Exits the Dead Player State
        /// </summary>
		public override void exit()
		{
			_time = 0;

			//Pick a random point out of the list of spawn points and respawn the player at that point if there is
			// no objects currently at that point, otherwise loop through the list of spawn points until you find an open
			// spawn point
			Random rand = new Random();
            int randVal = rand.Next(0, _player.PlayerManager.SpawnpointTimers.Count - 1);
			Geom geometryPlayerIsCollidingWith = PhysicsManager.getSingleton.PhysicsSimulator.Collide(_player.PlayerManager.RespawnPoints[randVal]);
			while (_player.PlayerManager.SpawnpointTimers[randVal] > 0.0 || geometryPlayerIsCollidingWith != null &&
				(geometryPlayerIsCollidingWith.CollisionCategories == Enums.CollisionCategories.Cat5 ||
				 geometryPlayerIsCollidingWith.CollisionCategories == Enums.CollisionCategories.Cat6))
            {
                ++randVal;
                if (randVal >= _player.PlayerManager.SpawnpointTimers.Count)
                    randVal = 0;
				geometryPlayerIsCollidingWith = PhysicsManager.getSingleton.PhysicsSimulator.Collide(_player.PlayerManager.RespawnPoints[randVal]);
            }

            Vector2 randPoint = _player.PlayerManager.RespawnPoints[randVal];
            _player.PlayerManager.startSpawnPointTimer(randVal);
			_player.HealthPoints = Player.MAX_HEALTH;
			_player.Entity.Active = true;
			_player.Entity.PhysicsBody.Body.Position = randPoint;
			_player.Entity.Sprite.Position = randPoint;

			Camera2D.getSingleton.register(_player.Entity.Name);
			foreach (SprayEmitter spray in _spawnEffect.Emitters)
			{
				spray.IsActive = true;
			}
		}

        /// <summary>
        /// Handles the collisions when something collides with the player while in the Dead Player State
        /// </summary>
        /// <param name="g1">The Player geometry</param>
        /// <param name="g2">The Geometry of the other object</param>
        /// <param name="cl">The lost of contact points</param>
        /// <returns></returns>
        public override bool collisionCallback(Geom g1, Geom g2, ContactList cl)
        {
			return true;
        }

        /// <summary>
        /// Handles the animation of the Player while in the Dead Player State
        /// </summary>
        /// <param name="animation"></param>
        public override void animationCallback(string animation)
        { }

		#endregion
	}
}
