using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using FarseerGames.FarseerPhysics;
using FarseerGames.FarseerPhysics.Dynamics;
using FarseerGames.FarseerPhysics.Collisions;

using rhodes.engine;
using rhodes.graphics;
using rhodes.graphics.particles;
using rhodes.graphics.particles.emitters;
using rhodes.input;
using rhodes.physics;
using sedohr.weapon;
using sedohr.objects;

namespace sedohr.player
{
	/// <summary>
	/// Represents a character in the game that can run and jump.
	/// Author: Robert Yates
	/// Modified: Anthony Reese, Josh Wilson, Brian Murphy
	/// Commented by: Andrew Kane and Robert Yates
	/// </summary>
	class Player
	{
		#region Constants

		/// <summary>
		/// The maximum speed of the player in the x direction
		/// </summary>
		public const float MAX_HORIZONTAL_PLAYER_VELOCITY = 300;

		/// <summary>
		/// The force used to move the player when on the ground
		/// </summary>
		public const float HORIZONTAL_PLAYER_FORCE_ON_GROUND = 15000;

		/// <summary>
		/// The force used to move the player when in the air
		/// </summary>
		public const float HORIZONTAL_PLAYER_FORCE_IN_AIR = 3000;

		/// <summary>
		/// The force applied to the player to make him/her jump.
		/// </summary>
		public const float PLAYER_JUMP_FORCE = -150000;

		/// <summary>
		/// The max amount of health a player can have.
		/// </summary>
		public const int MAX_HEALTH = 100;

		/// <summary>
		/// Offset for when making melee particle effect
		/// </summary>
		private const int MELEE_OFFSET = 25;

		/// <summary>
		/// The range for close quarters
		/// </summary>
		private const float CLOSE_QUARTER_RANGE = 100;

		/// <summary>
		/// The close quarters damage amount
		/// </summary>
		private const float CLOSE_QUARTER_MELEE_DAMAGE = 15;

		/// <summary>
		/// The close quarters cooldown
		/// </summary>
		private const float CLOSE_QUARTER_COOLDOWN = 2;

		#endregion

		#region Attributes

		/// <summary>
		/// The state of the player is currently in.
		/// Ex. Idle, Running, Jumping, Dead, Respawning
		/// </summary>
		private PlayerState _currentState;

		/// <summary>
		/// The object that will manage the dynamically added objects
		/// </summary>
        private ObjectManager _objectManager;

		/// <summary>
		/// A collection of all the states
		/// </summary>
		private Dictionary<String, PlayerState> _states;

		/// <summary>
		/// An instance of the class used to control the player's movements.
		/// </summary>
		private PlayerControl _playerControl;

		/// <summary>
		/// The unique number of this player. (1-4)
		/// </summary>
		private PlayerIndex _playerIndex;

		/// <summary>
		/// Layer for showing particle effects
		/// </summary>
		private static Layer _particleLayer;

		/// <summary>
		/// The particles for the melee attack
		/// </summary>
		private ParticleEngine _meleeParticle;

		/// <summary>
		/// Particle effect shown when a player dies.
		/// </summary>
		private ParticleEngine _deathEffect;

		/// <summary>
		/// The name of the player determined by the player index.
		/// Ex. If the player index is 1, the player's name is "player1"
		/// </summary>
		private String _name;

		/// <summary>
		/// The physics body to use for close quarters collisions
		/// </summary>
		private PhysicsBody _closeQuarterBody;

		/// <summary>
		/// The close quarters cooldown
		/// </summary>
		private float _closeCooldown;

		/// <summary>
		/// Whether the player is within close quarters or not
		/// </summary>
		private bool _closeFlag;

		/// <summary>
		/// The direction of melee
		/// </summary>
		private Vector2 meleeDirection;

		/// <summary>
		/// The sprite image that represents this player.
		/// </summary>
		private SimpleEntity _entity;

		/// <summary>
		/// Number of kills logged by the player.
		/// </summary>
		private int _kills;

		/// <summary>
		/// Stores information about last person to hit this player
		/// </summary>
		private string _lastHitBy;

		/// <summary>
		/// The current weapon the player is holding.
		/// </summary>
		private Weapon _currentWeapon;

		/// <summary>
		/// Whether to flip the sprite or not
		/// </summary>
		private Boolean _face;

		/// <summary>
		/// All the possible weapons the player can hold.
		/// </summary>
		private Dictionary<String, Weapon> _weapons;

		/// <summary>
		/// The manager that manages this player
		/// </summary>
		private PlayerManager _playerManager;

		#endregion

		#region Accessors

		/// <summary>
		/// Obtains the player's PlayerControl
		/// </summary>
		public PlayerControl PlayerControl
		{
			get
			{
				return _playerControl;
			}
		}

		/// <summary>
		/// Obtains the PlayerIndex (Player Number) of the player
		/// </summary>
		public PlayerIndex PlayerNumber
		{
			get
			{
				return _playerIndex;
			}
		}

		/// <summary>
		/// Gets or sets the engine to be used for the melee particle effects
		/// </summary>
		public ParticleEngine MeleeParticle
		{
			get { return _meleeParticle; }
			set { _meleeParticle = value; }
		}

		/// <summary>
		/// Obtains the player's name
		/// </summary>
		public string PlayerName
		{
			get
			{
				return _name;
			}
		}

		/// <summary>
		/// Gets the SimpleEntity of the player
		/// </summary>
		public SimpleEntity Entity
		{
			get
			{
				return _entity;
			}
		}

		/// <summary>
		/// Gets or sets the number of kills
		/// </summary>
		public int Kills
		{
			get
			{
				return _kills;
			}
			set
			{
				_kills = value;
			}
		}

		/// <summary>
		/// Number of deaths for the player.
		/// </summary>
		private int _deaths;
		public int Deaths { get { return _deaths; } set { _deaths = value; } }

		/// <summary>
		/// The amount of health the player contains.
		/// </summary>
		private float _healthPoints;
		public float HealthPoints
		{
			get { return _healthPoints; }
			set
			{
				_healthPoints = value;
				_healthPoints = MathHelper.Clamp(_healthPoints, 0, MAX_HEALTH);
			}
		}

		/// <summary>
		/// Gets the current weapon the player is using
		/// </summary>
		public Weapon CurrentWeapon
		{
			get { return _currentWeapon; }
		}

		/// <summary>
		/// Gets the dictionary of weapons
		/// </summary>
		public Dictionary<String, Weapon> Weapons
		{
			get { return _weapons; }
		}

		/// <summary>
		/// Gets the player manager managing this player
		/// </summary>
		public PlayerManager PlayerManager
		{
			get { return _playerManager; }
		}

		/// <summary>
		/// Gets or sets whether to flip the player sprite or not
		/// </summary>
		public Boolean Face
		{
			get
			{
				return _face;
			}
			set
			{
				_face = value;
			}
		}

		/// <summary>
		/// Gets the player's shoulder position
		/// </summary>
		public Vector2 ShoulderPosition
		{
			get
			{
				Vector2 shoulderPos = _entity.PhysicsBody.Body.Position;
				shoulderPos.Y = shoulderPos.Y - 20;

				if (_face)
				{
					shoulderPos.X = shoulderPos.X + 7;
				}
				else
				{
					shoulderPos.X = shoulderPos.X - 7;
				}

				return shoulderPos;
			}
		}

		#endregion

		#region Constructor

		/// <summary>
		/// Creates a player with a specific index
		/// </summary>
		/// <param name="playerIndex">The index of the player (1-4)</param>
		/// <param name="input">The input to control the player</param>
		/// <param name="objectManager">The object manager to handle the bare Player objects</param>
		/// <param name="playerManager">The player manager to handle the player's logic</param>
		public Player(PlayerIndex playerIndex, Input input, ObjectManager objectManager, PlayerManager playerManager)
		{
			
			// assign member values
			_playerIndex = playerIndex;
            _objectManager = objectManager;
            _playerManager = playerManager;

			_face = true;

			// cast the player index from an enum to an int and add 1 since casting returns (0-3)
			_name = "player" + ((int)playerIndex + 1);
			_playerControl = new PlayerControl(playerIndex, input);

			// create player entity
			_entity = EntityManager.getSingleton.loadSimple("assets/entities/" + _name);
			_entity.PhysicsBody.Body.InverseMomentOfInertia = 0f;
			foreach (Geom geom in _entity.PhysicsBody.GeomList)
			{
				geom.RestitutionCoefficient = 0; // prevent player from bouncing off floor
			
				geom.Tag = _name;
				geom.CollisionHandler += this.collisionHandler;

				// assign collision categories
				if (playerIndex == PlayerIndex.One)
				{
					geom.CollisionCategories = Enums.CollisionCategories.Cat1;
					geom.CollidesWith =
						Enums.CollisionCategories.Cat2 |
						Enums.CollisionCategories.Cat3 |
						Enums.CollisionCategories.Cat4 |
						Enums.CollisionCategories.Cat5 |
						Enums.CollisionCategories.Cat6;
				}
				else if (playerIndex == PlayerIndex.Two)
				{
					geom.CollisionCategories = Enums.CollisionCategories.Cat2;
					geom.CollidesWith =
						Enums.CollisionCategories.Cat1 |
						Enums.CollisionCategories.Cat3 |
						Enums.CollisionCategories.Cat4 |
						Enums.CollisionCategories.Cat5 |
						Enums.CollisionCategories.Cat6;
				}
				else if (playerIndex == PlayerIndex.Three)
				{
					geom.CollisionCategories = Enums.CollisionCategories.Cat3;
					geom.CollidesWith =
						Enums.CollisionCategories.Cat1 |
						Enums.CollisionCategories.Cat2 |
						Enums.CollisionCategories.Cat4 |
						Enums.CollisionCategories.Cat5 |
						Enums.CollisionCategories.Cat6;
				}
				else if (playerIndex == PlayerIndex.Four)
				{
					geom.CollisionCategories = Enums.CollisionCategories.Cat4;
					geom.CollidesWith =
						Enums.CollisionCategories.Cat1 |
						Enums.CollisionCategories.Cat2 |
						Enums.CollisionCategories.Cat3 |
						Enums.CollisionCategories.Cat5 |
						Enums.CollisionCategories.Cat6;
				}
			}
			_entity.Sprite._animationHandler = animationHandler;

			Camera2D.getSingleton.register(_entity.Sprite.Name);

			// populate states here
			_states = new Dictionary<string, PlayerState>();
			_states.Add("OnGround", new OnGroundPlayerState(this));
			_states.Add("InAir", new InAirPlayerState(this));
			_states.Add("Dead", new DeadPlayerState(this));
			_currentState = _states["OnGround"];

			// assign health points
			_healthPoints = 100;

			// populate weapons here (default to pistol)
			_weapons = new Dictionary<string, Weapon>();
			_weapons.Add("rocketLauncher", new RocketLauncher(playerIndex));
			_weapons.Add("laser", new Laser(playerIndex));
			_currentWeapon = new Pistol(playerIndex);
			_currentWeapon.Visible = true;
			_weapons.Add("pistol", _currentWeapon);


            // close quarter physics body for melee and for interacting with items
            Body closeBody = BodyFactory.Instance.CreateCircleBody(
                PhysicsManager.getSingleton.PhysicsSimulator, 10, 1);

			closeBody.Tag = "melee " + _name;
            List<Geom> closeGeomList = new List<Geom>();
            Geom closeGeom = GeomFactory.Instance.CreateCircleGeom(
                PhysicsManager.getSingleton.PhysicsSimulator, closeBody, CLOSE_QUARTER_RANGE, 10);

			closeGeom.Tag = "melee " + _name;
            closeGeom.CollisionCategories = _entity.PhysicsBody.GeomList[0].CollisionCategories;
            closeGeom.CollidesWith = Enums.CollisionCategories.Cat6;
            closeGeom.CollisionResponseEnabled = false;
            closeGeom.CollisionEnabled = false;
            closeGeom.CollisionHandler += this.closeQuarterHandler;
            closeGeomList.Add(closeGeom);
			_closeCooldown = 0.0f;
			_closeFlag = false;

            _closeQuarterBody = PhysicsManager.getSingleton.addPhysicsBody(
                "melee " + _playerIndex, closeBody, closeGeomList);

            //_pNameSprite = SpriteManager.getSingleton.load(_name + "name", "assets/sprites/" + _name + "name");
            //_pNameSprite.Scale = new Vector2(2.0f, 2.0f);

			//_meleeParticle = ParticleManager.getSingleton.load("melee", "assets/particles/meleeEffect");
			if (_particleLayer == null)
			{
			    _particleLayer = SceneManager.getSingleton.getActiveScene().getLayer(5);
			}
			_deathEffect = ParticleManager.getSingleton.load("death", "assets/particles/deathEffect");
			_particleLayer.AddParticle("death" + _entity.Name, _deathEffect);
			//foreach (PointEmitter pEmitter in _meleeParticle.Emitters)
			//{
			//    pEmitter.IsActive = false;
			//}
			//_particleLayer.AddParticle("melee" + _playerIndex, _meleeParticle);
		}

		#endregion

		#region Functions

		/// <summary>
		/// Function that is run when a collision occurs when this player
		/// </summary>
		/// <param name="g1">The geometry representing the player</param>
		/// <param name="g2">The geometry of something the player collided with</param>
		/// <param name="cl">The list of contacts</param>
		/// <returns>Returns true</returns>
		public bool collisionHandler(Geom g1, Geom g2, ContactList cl)
		{
			if (g2.CollisionCategories.Equals(Enums.CollisionCategories.Cat16))
				return false;

			string geom2 = g2.Tag.ToString();
			if (geom2.Equals("player1") || geom2.Equals("player2") || geom2.Equals("player3") || geom2.Equals("player4"))
				return false;

			if (g1.Tag.Equals(_name) && g2.Tag != null)
			{
				string collidedObjectName = g2.Tag.ToString();
                if (collidedObjectName.Contains("melee") && !collidedObjectName.Contains(_name))
                {
                    string enemy = collidedObjectName.Substring(collidedObjectName.LastIndexOf(' ') + 1);

                    if (_playerManager.Players[enemy]._closeQuarterBody.GeomList[0].CollisionEnabled)
                    {
                        if (!_playerManager.Players[enemy]._closeFlag)
                        {
                            if (takeDamage(CLOSE_QUARTER_MELEE_DAMAGE))
                            {
                                ProjectileManager.getSingleton.killed(enemy);
                            }

                            _playerManager.Players[enemy]._closeFlag = true;
                            _playerManager.Players[enemy]._closeQuarterBody.GeomList[0].CollisionEnabled = false;
                            _playerManager.Players[enemy]._closeQuarterBody.GeomList[0].CollisionResponseEnabled = false;

                            // knockback 
							_entity.PhysicsBody.Body.LinearVelocity = 500 * Vector2.Normalize(g1.Body.Position - _playerManager.Players[enemy].Entity.PhysicsBody.Body.Position);
                         }
                    }
                }

				if (collidedObjectName.Contains("projectile")
					&& !collidedObjectName.Contains(_name) && (_currentState.Name != "Dead"))
				{
					//Console.Out.WriteLine(_name + " hit by " + g2.Tag.ToString());
					_lastHitBy = collidedObjectName.Substring(0, collidedObjectName.IndexOf(' '));
					Projectile projectile = ProjectileManager.getSingleton.get(collidedObjectName);
					if (projectile != null && projectile.Active)
					{
						if (takeDamage(projectile.damage))
						{
							ProjectileManager.getSingleton.killed(collidedObjectName.
                                Substring(0, collidedObjectName.IndexOf(' ')));
						}
					}
					else // player was hit by a laser
					{
						if (takeDamage(LaserBeam.damage))
						{
							ProjectileManager.getSingleton.killed(collidedObjectName.
								Substring(0, collidedObjectName.IndexOf(' ')));
						}

					}
					
				} else if(collidedObjectName.Contains("splash")) {
					System.Diagnostics.Debug.WriteLine("SPLASH DAMAGE!");
				}
			}
			 
			return _currentState.collisionCallback(g1, g2, cl);
		}

		/// <summary>
		/// Function that is run when the player is within close quarters of a object
		/// </summary>
		/// <param name="g1">The geometry that is being hit</param>
		/// <param name="g2">The geometry that hit the first</param>
		/// <param name="cl">The list of contacts</param>
		/// <returns>Returns true</returns>
        public bool closeQuarterHandler(Geom g1, Geom g2, ContactList cl)
        {
			if (g2.CollisionCategories == Enums.CollisionCategories.Cat6)
			{
				String[] pair = ((String)g2.Body.Tag).Split('/');

				if (pair.Length != 2)
				{
					Vector2 distance = (g2.Body.Position - ShoulderPosition);
					distance.Normalize();

					SimpleEntity entity = _objectManager.getSimple((String)g2.Body.Tag).SimpleEntity;

					float rotation = _playerControl.getGunAngleOfRotation(ShoulderPosition);
					meleeDirection.X = (float)Math.Cos(rotation);
					meleeDirection.Y = (float)Math.Sin(rotation);

					if (Vector2.Dot(distance, meleeDirection) > 0)
					{
						entity.PhysicsBody.Body.LinearVelocity = meleeDirection * g2.Body.Mass * 60.0f;
						//entity.PhysicsBody.Body.ApplyImpulse(meleeDirection * g2.Body.Mass * 500.0f);
						_closeQuarterBody.GeomList[0].CollisionEnabled = false;
					}
				}
			}
            
            return true;
        }

		/// <summary>
		/// The callback that is run when the animation is switched
		/// </summary>
		/// <param name="animation">The name of the animation</param>
		public void animationHandler(String animation)
		{
			_currentState.animationCallback(animation);
			if (animation.Equals("kick"))
			{
				foreach (Geom geom in _closeQuarterBody.GeomList)
				{
					geom.CollisionEnabled = false;
					geom.CollisionResponseEnabled = false;
				}
			}
		}

		/// <summary>
		/// The player takes damage and dies if healthpoints go below 0
		/// </summary>
		/// <param name="damage">The amount of damage the player will take</param>
		/// <returns>Whether the player is dead or not</returns>
		public bool takeDamage(float damage)
		{
			if (_healthPoints - damage <= 0){
                // Huge rumble!
                PlayerControl.rumble(1.0f);
				foreach (PointEmitter pEmitter in _deathEffect.Emitters)
				{
					pEmitter.EmitterPosition = _entity.Sprite.Position;
					pEmitter.IsActive = true;
				}
				changeState("Dead");
				_deaths++;
				//_healthPoints = MAX_HEALTH;
				return true;
			}else{
                float rumble = (damage / 10.0f);
                MathHelper.Clamp(rumble, 0.1f, 1.0f);
                PlayerControl.rumble(rumble);
				_healthPoints -= damage;
				return false;
			}
		}

		/// <summary>
		/// Updates the player's movement
		/// </summary>
		public void update(float elapsedTime)
		{
			if (_playerControl.cooldown >= 0)
			{
				_playerControl.cooldown -= elapsedTime;
			}

			_currentState.update(elapsedTime);

			if (_currentWeapon.Genre == Weapon.weaponType.LASER && _playerControl.pushingFire()
				&& (_playerControl.cooldown <= 0))
			{
				_playerControl.LaserDuration -= elapsedTime;
			}


			bool pressingFire = (_playerControl.cooldown > 0 || _playerControl.LaserDuration < 0) ? false : _playerControl.pushingFire();

			if ((_entity.PhysicsBody.Body.Position.X > 3000 || _entity.PhysicsBody.Body.Position.X < -500 ||
				_entity.PhysicsBody.Body.Position.Y > 2500 || _entity.PhysicsBody.Body.Position.Y < -500) &&
				_currentState.Name != "Dead")
			{
				//Console.Out.WriteLine("Player out of bounds! DEAD!!!");
				_deaths += 1;
				if (_lastHitBy != null)
				{
					ProjectileManager.getSingleton.killed(_lastHitBy);
				}
				foreach (PointEmitter pEmitter in _deathEffect.Emitters)
				{
					pEmitter.EmitterPosition = _entity.Sprite.Position;
					pEmitter.IsActive = true;
				}
				changeState("Dead");

				//Possibly add a death effect to cause the player to be sucked into the big black hole in the middle
			}


			// get the rotation of the weapon
			float rotation = _playerControl.getGunAngleOfRotation(ShoulderPosition);
		
			if (_currentState.Name != "Dead")
			{

				foreach (Weapon weapon in Weapons.Values)
				{
					bool pressingWeaponButtonOfThisWeapon = false;

					if (weapon == _currentWeapon)
					{
						pressingWeaponButtonOfThisWeapon = pressingFire;
                        if (_currentWeapon.Genre != Weapon.weaponType.LASER)
                        {
                            if (pressingFire)
                                _playerControl.cooldown = weapon.cooldown;
                        }
                        else
                        {
                            if (_playerControl.LaserDuration < 0)
                            {
                                _playerControl.cooldown = weapon.cooldown;
                                _playerControl.LaserDuration = weapon.WeaponDuration;
                            }
                            //if (!_playerControl.pushingFire() && _playerControl.hasJustPushedFireButton())
                            //    _playerControl.cooldown = weapon.cooldown;
                            if (!pressingFire)
                            {
                                _playerControl.LaserDuration = weapon.WeaponDuration;
                            }
                        }
					}

					weapon.update(elapsedTime, pressingWeaponButtonOfThisWeapon, ShoulderPosition, rotation);
				}

				if (pressingFire == false)
				{
					sedohr.weapon.Weapon.weaponType wType = _playerControl.weaponButtonDown();
					switch (wType)
					{
						case Weapon.weaponType.PISTOL:
							if (_currentWeapon.Genre != Weapon.weaponType.PISTOL)
							{
								_currentWeapon.Visible = false;
								_currentWeapon = _weapons["pistol"];
								_currentWeapon.Visible = true;
								_playerControl.cooldown = _currentWeapon.cooldown/4;
								_playerControl.LaserDuration = _currentWeapon.WeaponDuration;
							}
							break;
						case Weapon.weaponType.LASER:
							if (_currentWeapon.Genre != Weapon.weaponType.LASER)
							{
								_currentWeapon.Visible = false;
								_currentWeapon = _weapons["laser"];
								_currentWeapon.Visible = true;
								_playerControl.cooldown = _currentWeapon.cooldown/4;
								_playerControl.LaserDuration = _currentWeapon.WeaponDuration;
							}
							break;
						case Weapon.weaponType.ROCKET:
							if (_currentWeapon.Genre != Weapon.weaponType.ROCKET)
							{
								_currentWeapon.Visible = false;
								_currentWeapon = _weapons["rocketLauncher"];
								_currentWeapon.Visible = true;
								_playerControl.cooldown = _currentWeapon.cooldown/4;
								_playerControl.LaserDuration = _currentWeapon.WeaponDuration;
							}
							break;
					}
				}

				if ((rotation > 1.6f) || (rotation < -1.6f)) //MathHelper.PiOver2 + 0.02f )
				{
					_entity.Sprite.flipLeft();
					_currentWeapon.Sprite.flipUp();
					_face = false;
				}
				else if((rotation < 1.5f) && (rotation > -1.5f)) //MathHelper.PiOver2) && !_face )
				{	
					_entity.Sprite.flipRight();
					_currentWeapon.Sprite.flipDown();
					_face = true;
				}
			}
			if (_closeFlag)
			{
				_closeCooldown += elapsedTime;
				if (_closeCooldown >= CLOSE_QUARTER_COOLDOWN)
				{
					_closeCooldown = 0.0f;
					_closeFlag = false;
				}
			}
            _closeQuarterBody.Body.Position = _entity.PhysicsBody.Body.Position;
		}

		/// <summary>
		/// Handles the player's input
		/// </summary>
		public void handle()
		{
			if (_playerControl.hasJustPushedMeleeButton())
			{
				if (!_closeFlag)
				{
					_entity.Sprite.setAnimation("kick", false);
					_closeQuarterBody.GeomList[0].CollisionEnabled = true;
					Vector2 offset = Vector2.Zero;
					if (this._face)
					{
						offset.X += MELEE_OFFSET;
					}
					else
					{
						offset.X -= MELEE_OFFSET;
					}
					//foreach (PointEmitter pEmitter in _meleeParticle.Emitters)
					//{
					//    pEmitter.EmitterPosition = _closeQuarterBody.Position + offset;
					//    pEmitter.IsActive = true;
					//}
				}
			}
            
			_currentState.handle();
		}

		/// <summary>
		/// Changes the player's state.
		/// </summary>
		/// <param name="stateName">The name of the new state</param>
		public void changeState(String stateName)
		{
			_lastHitBy = null;

			// find the state
			if (!_states.ContainsKey(stateName))
				return;

			// exit current state
			_currentState.exit();

			// change state
			_currentState = _states[stateName];

			// enter new state
			_currentState.enter();
		}
		#endregion

	}
}
