/* \class PlayerControl
 * \class Input
 * Input.cs
 * 
 * \author Andrew Kane
 *
 *  Defines how the player will interface with the input
 */

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using rhodes.input;
using Microsoft.Xna.Framework.Input;
using sedohr.weapon;
using System.Timers;

namespace sedohr.player
{
    /// <summary>
    /// Class that will deal with the input for a player
	/// Author: Andrew Kane
    /// </summary>
    class PlayerControl
	{
		#region Attributes

		/// <summary>
        /// Which number the player is
        /// </summary>
        private PlayerIndex playerNum;

        /// <summary>
        /// Input system being used for the players
        /// </summary>
        private Input input;

		/// <summary>
		/// The current rotation of the weapon
		/// </summary>
		private float rot;

		/// <summary>
		/// The time left until the player can fire again (in miliseconds)
		/// </summary>
		private float cooldownCounter;

		/// <summary>
		/// The length of time the laser is being fired
		/// </summary>
		private float laserDuration;

        /// <summary>
        /// Timer used for disabling rumble
        /// </summary>
        private Timer timer;

		#endregion

		#region Constructor

		/// <summary>
		/// PlayerControl Constructor
		/// </summary>
		/// <param name="playerIndex">Which player number we are monitoring</param>
		/// <param name="_input">Which input system to use for the players</param>
		public PlayerControl(PlayerIndex playerIndex, Input _input)
		{
			playerNum = playerIndex;
			input = _input;
            timer = new Timer(500);
            timer.Elapsed += new ElapsedEventHandler(stopRumble);
            
			return;
		}

		#endregion

		#region Accessors

		/// <summary>
		/// Gets or sets the Player's rotation
		/// </summary>
		public float rotation
		{
			get
			{
				return rot;
			}
			set
			{
				rot = value;
			}
		}

		/// <summary>
		/// Gets or sets the player's current cooldown
		/// </summary>
		public float cooldown
		{
			get
			{
				return cooldownCounter;
			}
			set
			{
				cooldownCounter = value;
			}
		}

        /// <summary>
        /// Gets or sets the player's current duration of the laser
        /// </summary>
        public float LaserDuration
        {
            get
			{
				return laserDuration;
			}
            set
			{
				laserDuration = value;
			}
		}

		#endregion

		#region Functions

        /// <summary>
        /// Makes the controller rumble
        /// </summary>
        public void rumble(float amount)
        {
            input.current[(int)playerNum].rumble(amount, amount);
            timer.Start();
            return;
        }

        /// <summary>
        /// Stops the controller's rumble
        /// </summary>
        public void stopRumble(object sender, ElapsedEventArgs e)
        {
            input.current[(int)playerNum].rumble(0, 0);
            timer.Stop();
            return;
        }

		/// <summary>
        /// Determines which angle (in radians) the gun/weapon is to be rotated to
        /// </summary>
        /// <param name="shoulderPos">The position of the player's shoulder</param>
        /// <returns>The angle (in radians) to rotate the gun to</returns>
        public float getGunAngleOfRotation(Vector2 shoulderPos)
        {
			//System.Diagnostics.Debug.WriteLine("zomg aiming " + input.current[(int)playerNum].connected);
#if !XBOX
            if(input.current[(int)playerNum].useKeyboard)
            {
				//System.Diagnostics.Debug.WriteLine("keyboard!");
                Vector2 mousePos = input.current[(int)playerNum].mousePosition;
                float dx = mousePos.X - shoulderPos.X,
                    dy = mousePos.Y - shoulderPos.Y;
                return (float)(Math.Atan2(dy, dx));
            }
            else if(input.current[(int)playerNum].connected)
#else
            if(input.current[(int)playerNum].connected)
#endif
            {
				if (Math.Sqrt(Math.Pow(input.current[(int)playerNum].aimVertical, 2) + Math.Pow(input.current[(int)playerNum].aimHorizontal, 2))
					>= input.current[(int)playerNum].thumbstickSens)
				{
					rotation = (float)(Math.Atan2(-input.current[(int)playerNum].aimVertical,
						input.current[(int)playerNum].aimHorizontal));
				}
				return rotation;
            }
            return 0;
        }

        /// <summary>
        /// Checks whether the player is attempting to move the player left or not
        /// </summary>
        /// <returns>Whether the player is attempting to move left or not</returns>
        public bool isPushingLeft()
        {
            return (input.current[(int)playerNum].left == ButtonState.Pressed);
        }

        /// <summary>
        /// Gets the amount of force the player is providing to the left movement (0 ~ 1)
        /// </summary>
        /// <returns>0 ~ 1 of the amount of force being pushed left</returns>
        public float getLeftPush()
        {
			if (input.current[(int)playerNum].left == ButtonState.Pressed)
				return 1.0f;
            return (input.current[(int)playerNum].moveHorizontal < 0.0f) ?
                -input.current[(int)playerNum].moveHorizontal : 0;
        }

        /// <summary>
        /// Checks whether the player is attempting to move the player right or not
        /// </summary>
        /// <returns>Whether the player is attempting to move right or not</returns>
        public bool isPushingRight()
        {
            return (input.current[(int)playerNum].right == ButtonState.Pressed);
        }

        /// <summary>
        /// Gets the amount of force the player is providing to the right movement (0 ~ 1)
        /// </summary>
        /// <returns>0 ~ 1 of the amount of force being pushed right</returns>
        public float getRightPush()
        {
			if (input.current[(int)playerNum].right == ButtonState.Pressed)
				return 1.0f;
            return (input.current[(int)playerNum].moveHorizontal > 0.0f) ?
                input.current[(int)playerNum].moveHorizontal : 0;
        }

        /// <summary>
        /// Whether the player has just pushed the jump button (and hot held it down)
        /// </summary>
        /// <returns>If the user just pushed the jump button</returns>
        public bool hasJustPushedJumpButton()
        {
            //return (input.current[(int)playerNum].jump == ButtonState.Pressed &&
            //    input.previous[(int)playerNum].jump == ButtonState.Released);
			return input.current[(int)playerNum].jump == ButtonState.Pressed;
        }

        /// <summary>
        /// Whether the player has just pushed the melee button (and hot held it down)
        /// </summary>
        /// <returns>If the user just pushed the melee button</returns>
        public bool hasJustPushedMeleeButton()
        {
            return (input.current[(int)playerNum].melee == ButtonState.Pressed &&
                input.previous[(int)playerNum].melee == ButtonState.Released);
        }

        /// <summary>
        /// Whether the player has just pushed the fire button
        /// </summary>
        /// <returns>If the user just pushed the fire button</returns>
        public bool hasJustPushedFireButton()
        {
            return input.current[(int)playerNum].fire == ButtonState.Pressed
                && input.previous[(int)playerNum].fire == ButtonState.Released;
        }

        /// <summary>
        /// Whether the player is pushing the fire button
        /// </summary>
        /// <returns>If the user is pushing the fire button</returns>
        public bool pushingFire()
        {
            return (input.current[(int)playerNum].fire == ButtonState.Pressed);
        }

        /// <summary>
        /// Whether the player has pressed a weapon button and if so, which weapon type it is
        /// </summary>
        /// <returns>The weapon type that the user pressed the button of</returns>
        public Weapon.weaponType weaponButtonDown()
        {
            //if(input.current[(int)playerNum].gun == ButtonState.Pressed && input.previous[(int)playerNum].gun == ButtonState.Released)
            if(input.current[(int)playerNum].gun == ButtonState.Pressed)
            {
                return Weapon.weaponType.PISTOL;
            }
            //else if(input.current[(int)playerNum].laser == ButtonState.Pressed && input.previous[(int)playerNum].laser == ButtonState.Released)
            else if(input.current[(int)playerNum].laser == ButtonState.Pressed)
            {
                return Weapon.weaponType.LASER;
            }
            //else if(input.current[(int)playerNum].missile == ButtonState.Pressed && input.previous[(int)playerNum].missile == ButtonState.Released)
            else if(input.current[(int)playerNum].missile == ButtonState.Pressed)
            {
                return Weapon.weaponType.ROCKET;
            }
            else
            {
                return Weapon.weaponType.NONE;
            }
        }

		#endregion
	}
}
