using System;
using System.Collections.Generic;
using System.Text;
using rhodes.physics;
using Microsoft.Xna.Framework;
using FarseerGames.FarseerPhysics.Collisions;

namespace sedohr.weapon
{
    /// <summary>
    /// Class that manages the explosions
	/// Author: Andrew Kane
    /// </summary>
    class ExplosionManager
	{
		#region Attributes

		/// <summary>
        /// Static instance of the ProjectileManager class so that we can use it as a singleton
        /// </summary>
        private static volatile ExplosionManager instance;

        /// <summary>
        /// Object to sync/lock around to make sure nothing is re-initialized
        /// </summary>
        private static object syncRoot = new Object();

        /// <summary>
        /// A list of all the projectiles
        /// </summary>
        private SortedDictionary<String, Explosion> explosions;

        /// <summary>
        /// An array that contains all of the splash area physics bodies (not sure why we need physics bodies for these, but...)
        /// </summary>
        protected Dictionary<string, PhysicsBody> splashBodies;

		/// <summary>
		/// Listing of explosions to remove
		/// </summary>
		private List<string> removeExplosions;

		#endregion

		#region Constructor

		/// <summary>
        /// Default constructor for ExplosionManager
        /// </summary>
        public ExplosionManager()
        {
            explosions = new SortedDictionary<string, Explosion>();
            splashBodies = new Dictionary<string, PhysicsBody>();
			removeExplosions = new List<string>();
            return;
		}

		#endregion

		#region Accessors

		/// <summary>
        /// Returns a ExplosionManager object that is a singleton
        /// </summary>
        public static ExplosionManager getSingleton
        {
            get
            {
                if(instance == null)
                {
                    lock(syncRoot)
                    {
                        if(instance == null)
                            instance = new ExplosionManager();
                    }
                }
                return instance;
            }
        }

		#endregion

		#region Functions

		/// <summary>
        /// Creates a new explosion
        /// </summary>
        /// <param name="position">The position of the explosion</param>
        /// <param name="splash">The radius ("splash size") of the explosion</param>
        /// <param name="funct">The collision handler delegate</param>
        /// <returns></returns>
        public Explosion createExplosion(Vector2 position, float splash, Geom.CollisionHandlerDelegate funct)
        {
            Explosion expl = new Explosion(splashBodies, position, splash, funct);
            explosions.Add(expl.name, expl);
            return expl;
        }

        /// <summary>
        /// Removes an explosion
        /// </summary>
        /// <param name="expl">The explosion object</param>
        public void remove(Explosion expl)
        {
            deactivate(expl.name);
            explosions.Remove(expl.name);
            return;
        }

        /// <summary>
        /// Removes an explosion
        /// </summary>
        /// <param name="name">The name of the explosion</param>
        public void remove(string name)
        {
            deactivate(name);
			removeExplosions.Add(name);
            //explosions.Remove(name);
            return;
        }

        /// <summary>
        /// Deactivates an explosion
        /// </summary>
        /// <param name="name">The name of the explosion</param>
        protected void deactivate(string name)
        {
            // Do any deactivation shenanigans here
            splashBodies[name].Body.Enabled = false;
            return;
        }

        /// <summary>
        /// Updates all of the explosions
        /// </summary>
        /// <param name="elapsedTime">Time that elapsed</param>
        public void update(long elapsedTime)
        {
			foreach(Explosion expl in explosions.Values)
                expl.update(elapsedTime);
			foreach (String name in removeExplosions)
				explosions.Remove(name);
			removeExplosions.Clear();

            return;
		}

		#endregion

	}
}
