using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;

using FarseerGames.FarseerPhysics;
using FarseerGames.FarseerPhysics.Dynamics;
using FarseerGames.FarseerPhysics.Collisions;

using rhodes.engine;
using rhodes.graphics;
using rhodes.graphics.particles;
using rhodes.graphics.particles.emitters;
using rhodes.physics;
using rhodes.audio;

using sedohr.player;

/*
 * Projectile.cs
 * A projectile is a bullet or a rocket that can be shot by a weapon.
 * 
 * Created By: Andrew Kane
 * Modified & Updated by: Rob Yates, Anthony Reese, Josh Wilson
 * Commented By: Rob Yates
 */
namespace sedohr.weapon
{
	/// <summary>
	/// This class represents a projectile (a bullet or a rocket) that
	/// can be shot by a weapon. Projectile inherits from
	/// rhodes.engine.SimpleEntity which contains a sprite and physics body.
	/// </summary>
	class Projectile : SimpleEntity
	{
		#region Constants

		/// <summary>
		/// The number of explosions
		/// </summary>
		private const int _NUMOFEXPLOSIONS = 16;

		/// <summary>
		/// The number of explosions rocket trails
		/// </summary>
		private const int _NUMOFTRAILS = 16;

		/// <summary>
		/// The maximum x value of a projectile's position before it is out of bounds
		/// </summary>
		private const int _MAXIMUM_X_BOUNDS = 3000;

		/// <summary>
		/// The minimum x value of a projectile's position before it is out of bounds
		/// </summary>
		private const int _MINIMUM_X_BOUNDS = -500;

		/// <summary>
		/// The maximum y value of a projectile's position before it is out of bounds
		/// </summary>
		private const int _MAXIMUM_Y_BOUNDS = 2500;

		/// <summary>
		/// The minimum y value of a projectile's position before it is out of bounds
		/// </summary>
		private const int _MINIMUM_Y_BOUNDS = -500;


		/// <summary>
		/// A string with the text pistol
		/// </summary>
		private const string PISTOL_STRING = "pistol";

		/// <summary>
		/// A string with the text rocket
		/// </summary>
		private const string ROCKET_STRING = "rocket";

		/// <summary>
		/// A string with the text rocket
		/// </summary>
		private const string LASER_STRING = "laser";

		#region Rocket-Specific Constants

		/// <summary>
		/// The relative velocity between the expelled fuel and the rocket
		/// </summary>
		private const int RELATIVE_VELOCITY = 375;
		/// <summary>
		/// The mass of the rocket when it is "empty" without fuel
		/// </summary>
		private const int ROCKET_MIN_MASS = 5;
		/// <summary>
		/// The rate at which rocket fuel burns (in mass/time)
		/// </summary>
		private const int FUEL_BURN_RATE = 20;
		/// <summary>
		/// The force continuously applied to the rocket in mid-air.
		/// (When the rocket runs out of fuel there is no more thrust.)
		/// </summary>
		private const int ROCKET_THRUST = FUEL_BURN_RATE * RELATIVE_VELOCITY;

		/// <summary>
		/// The radius of a rocket's splash damage
		/// </summary>
		private const int ROCKET_SPLASH = 10;

		/// <summary>
		/// The amount of force to initially shoot the rocket
		/// </summary>
		private const int _ROCKET_FORCE_STRENGTH = 1200000;

		/// <summary>
		/// The initial mass of the rocket
		/// </summary>
		private const int _ROCKET_INITIAL_MASS = 30;

		/// <summary>
		/// The damage of the rocket
		/// </summary>
		private const int _ROCKET_DAMAGE = 30;

		#endregion

		#region Pistol-Specific Constants

		/// <summary>
		/// The amount of force to initially shoot the pistol bullet
		/// </summary>
		private const int _PISTOL_FORCE_STRENGTH = 500000;

		/// <summary>
		/// The initial mass of the pistol bullet
		/// </summary>
		private const int _PISTOL_INITIAL_MASS = 5;

		/// <summary>
		/// The damage of the pistol bullet
		/// </summary>
		private const int _PISTOL_DAMAGE =7;

		#endregion

		#endregion

		#region Attributes

		/// <summary>
		/// Layer needed for making particle effects show
		/// </summary>
		private static Layer _particleLayer;

		/// <summary>
		/// The index of the explosion particles
		/// </summary>
		private static int _explosionIndex = 0;

		/// <summary>
		/// The index of the rocket trail particles
		/// </summary>
		private static int _trialIndex = 0;

		/// <summary>
		/// The particles for explosions
		/// </summary>
		private static ParticleEngine[] _explosionParticles = new ParticleEngine[_NUMOFEXPLOSIONS];

		/// <summary>
		/// The particles for rocket trails
		/// </summary>
		private static ParticleEngine[] _trailParticles = new ParticleEngine[_NUMOFTRAILS];

		/// <summary>
		/// The position where projectiles go when they are deactivated
		/// </summary>
		private Vector2 OFF_SCREEN_POSITION = new Vector2(-5000.0f, -5000.0f);

		/// <summary>
		/// A vector to store this projectile's rotation
		/// </summary>
		Vector2 rotationVector = new Vector2();

		/// <summary>
		/// The name of the player that fired this projectile
		/// </summary>
		private string playerName;

		/// <summary>
		/// The player who shot this projectile
		/// </summary>
		protected PlayerIndex playerNum;

		/// <summary>
		/// The type of weapon that fired the projectile
		/// </summary>
		protected Weapon.weaponType weaponGenre;

		#endregion

		#region Accessors

		/// <summary>
		/// Gets the amount of damage the projectile will inflict on the player
		/// </summary>
		public float damage
		{
			get
			{
				if (weaponGenre.Equals(Weapon.weaponType.ROCKET))
				{
					return _ROCKET_DAMAGE;
				}
				else //if (weaponGenre.Equals(weaponType.PISTOL)) {
				{
					return _PISTOL_DAMAGE;
				}
			}
		}

		/// <summary>
		/// Gets the acceleration of the projectile and 
		/// reduce it for knocking back another player
		/// </summary>
		public Vector2 KnockBackAcceleration
		{
			get
			{
				Vector2 acceleration = (_physicsBody.Body.LinearVelocity *
					_physicsBody.Body.LinearVelocity) /
					2 * (_physicsBody.Body.Position - _initialPosition);
				return acceleration / 30000;
			}
		}

		#endregion

		#region Constructor

		/// <summary>
		/// Create the projectile, but don't use it until use method is called.
		/// </summary>
		/// <param name="projectileName">The unique name of this projectile</param>
		/// <param name="physicsBody">The physics body used for collision detection</param>
		/// <param name="projectileSprite">The sprite image of this projectile</param>
		public Projectile(String projectileName, PhysicsBody physicsBody, Sprite projectileSprite)
			:
			  base(projectileName, physicsBody, projectileSprite) { }

		#endregion

		#region Functions

		/// <summary>
		/// Utilize the projectile and assign the projectile's variables.
		/// </summary>
		/// <param name="projectileName">The unique name of this projectile</param>
		/// <param name="playerIndex">What player fired this projectile</param>
		/// <param name="genre">The type of weapon this projectile from</param>
		/// <param name="position">What position this projectile is starting at</param>
		/// <param name="rotation">What rotation this projectile is starting at</param>
		public void use(string projectileName, PlayerIndex playerIndex, Weapon.weaponType genre,
			Vector2 position, float rotation)
		{
			// reset physics body
			_physicsBody.Body.ClearForce();
			_physicsBody.Body.ClearTorque();
			_physicsBody.Body.LinearVelocity = Vector2.Zero;

			// reassign name
			_name = projectileName;
			_physicsBody.Body.Tag = projectileName;

			// set initial values
			playerNum = playerIndex;
			playerName = "player" + ((int)playerIndex + 1);
			weaponGenre = genre;

			// set to be active
			Active = true;

			// Set up initial particle layer if not set
			if (_particleLayer == null)
			{
				_particleLayer = SceneManager.getSingleton.getActiveScene().getLayer(5);
			}

			// set position, rotation, and splash radius
			_physicsBody.Body.Position = _sprite.Position = position;
			_initialPosition = position;
			_physicsBody.Body.Rotation = _sprite.Rotation = rotation;

			// set rotation vector
			rotationVector.X = (float)Math.Cos(_physicsBody.Body.Rotation);
			rotationVector.Y = (float)Math.Sin(_physicsBody.Body.Rotation);

			// apply initial mass and force strength depending on weapon type
			if (weaponGenre.Equals(Weapon.weaponType.ROCKET))
			{
				_physicsBody.Body.Mass = _ROCKET_INITIAL_MASS;
				_physicsBody.Body.ApplyForce(_ROCKET_FORCE_STRENGTH * rotationVector);
			}
			else //if (weaponGenre.Equals(Weapon.weaponType.PISTOL))
			{
				_physicsBody.Body.Mass = _PISTOL_INITIAL_MASS;
				_physicsBody.Body.ApplyForce(_PISTOL_FORCE_STRENGTH * rotationVector);
			}

			// set geometry tags and collision handler
			foreach (Geom geom in _physicsBody.GeomList)
			{
				geom.Tag = _name;
				geom.CollisionHandler = this.projectileAnythingCollisionHandler;
				if (playerIndex == PlayerIndex.One)
				{
					geom.CollisionCategories = Enums.CollisionCategories.Cat1;
					geom.CollidesWith =
						Enums.CollisionCategories.Cat2 |
						Enums.CollisionCategories.Cat3 |
						Enums.CollisionCategories.Cat4 |
						Enums.CollisionCategories.Cat5 |
						Enums.CollisionCategories.Cat6;
				}
				else if (playerIndex == PlayerIndex.Two)
				{
					geom.CollisionCategories = Enums.CollisionCategories.Cat2;
					geom.CollidesWith =
						Enums.CollisionCategories.Cat1 |
						Enums.CollisionCategories.Cat3 |
						Enums.CollisionCategories.Cat4 |
						Enums.CollisionCategories.Cat5 |
						Enums.CollisionCategories.Cat6;
				}
				else if (playerIndex == PlayerIndex.Three)
				{
					geom.CollisionCategories = Enums.CollisionCategories.Cat3;
					geom.CollidesWith =
						Enums.CollisionCategories.Cat1 |
						Enums.CollisionCategories.Cat2 |
						Enums.CollisionCategories.Cat4 |
						Enums.CollisionCategories.Cat5 |
						Enums.CollisionCategories.Cat6;
				}
				else if (playerIndex == PlayerIndex.Four)
				{
					geom.CollisionCategories = Enums.CollisionCategories.Cat4;
					geom.CollidesWith =
						Enums.CollisionCategories.Cat1 |
						Enums.CollisionCategories.Cat2 |
						Enums.CollisionCategories.Cat3 |
						Enums.CollisionCategories.Cat5 |
						Enums.CollisionCategories.Cat6;
				}
			}

			// apply particle effects to rocket
			if (weaponGenre.Equals(Weapon.weaponType.ROCKET))
			{
				if (_trailParticles[_trialIndex] == null)
				{
					_trailParticles[_trialIndex] =
						ParticleManager.getSingleton.load("trial",
						"assets/particles/trailEffect");
					_particleLayer.AddParticle("trail" + _trialIndex,
						_trailParticles[_trialIndex]);
				}
				_trailParticles[_trialIndex].reRegister(projectileName);
				foreach (PointEmitter pEmitter in _trailParticles[_trialIndex].Emitters)
				{
					pEmitter.EmitterPosition = position;
					pEmitter.IsActive = true;
				}
				if (_trialIndex < _NUMOFTRAILS - 1)
				{
					_trialIndex++;
				}
				else
				{
					_trialIndex = 0;
				}
			}
		}

		/// <summary>
		/// Updates the properties of the projectile
		/// </summary>
		/// <param name="time">The amount of time that has passed</param>
		public override void update(float elapsedTime)
		{
			// don't update if this projectile is not active
			if (!Active)
			{
				return;
			}

			// deactivate projectile if it is out of bounds
			if (_physicsBody.Body.Position.X > _MAXIMUM_X_BOUNDS ||
				_physicsBody.Body.Position.X < _MINIMUM_X_BOUNDS ||
				_physicsBody.Body.Position.Y > _MAXIMUM_Y_BOUNDS ||
				_physicsBody.Body.Position.Y < _MINIMUM_Y_BOUNDS)
			{
				deactivateProjectile();
			}
			else
			{
				// update physics body
				_physicsBody.Body.Rotation =
					(float)Math.Atan2(_physicsBody.Body.LinearVelocity.Y,
									  _physicsBody.Body.LinearVelocity.X);

				base.update(elapsedTime);
				// elapsed time is about 0.016 for 60 frames per seoond

				// if this projectile is a rocket
				if (weaponGenre.Equals(Weapon.weaponType.ROCKET))
				{
					//if rocket has fuel to burn
					if (_physicsBody.Body.Mass > ROCKET_MIN_MASS)
					{
						// set rotation vector
						rotationVector.X = (float)Math.Cos(_physicsBody.Body.Rotation);
						rotationVector.Y = (float)Math.Sin(_physicsBody.Body.Rotation);

						// apply rocket thrust
						_physicsBody.Body.ApplyForce(ROCKET_THRUST * rotationVector);

						// and decrease mass (implying fuel was burnt)
						_physicsBody.Body.Mass -= (float)(FUEL_BURN_RATE * elapsedTime);
					}
				}
			}
		}

		/// <summary>
		/// Handles collision when the projectile hits something
		/// </summary>
		/// <param name="g1">The projectile geometry</param>
		/// <param name="g2">The geometry of another object</param>
		/// <param name="cl">The list of contact points</param>
		/// <returns>True, if the projectile hit something. False, if not.</returns>
		public bool projectileAnythingCollisionHandler(Geom g1, Geom g2, ContactList cl)
		{
			// make it so projectiles cannot collide with splash damage physics bodies
			if (!g1.Tag.ToString().Contains("player") && g2.Tag.ToString().Contains("splash"))
				return false;

			// make it so projectiles cannot collide with black holes
			if (g2.CollisionCategories.Equals(Enums.CollisionCategories.Cat16))
				return false;

			// if the projectile hit something
			if (g1.Tag.Equals(_name) && g2.Tag != null)
			{

				// check hierarchy for weapons that shoot projectile
				string geom1 = null;
				string geom2 = g2.Tag.ToString();
				if (geom2.Contains(PISTOL_STRING) ||
					geom2.Contains(ROCKET_STRING) ||
					geom2.Contains(LASER_STRING))
				{
					if (weaponGenre.Equals(Weapon.weaponType.PISTOL))
						geom1 = PISTOL_STRING;
					else if (weaponGenre.Equals(Weapon.weaponType.ROCKET))
						geom1 = ROCKET_STRING;
					else if (weaponGenre.Equals(Weapon.weaponType.LASER))
						geom1 = LASER_STRING;

					if (geom2.Contains(PISTOL_STRING))
						geom2 = geom2.Substring(geom2.IndexOf(PISTOL_STRING), 6);
					else if (geom2.Contains(ROCKET_STRING))
						geom2 = geom2.Substring(geom2.IndexOf(ROCKET_STRING), 6);
					else
						geom2 = geom2.Substring(geom2.IndexOf(LASER_STRING), 5);

					// projectiles of a greater level will not be destroyed
					// upon collision. For example, a rocket will destroy
					// pistol bullets, but will not be destroyed itself.
					if ((int)ProjectileManager.getSingleton.WeaponLevels[geom1] >
						(int)ProjectileManager.getSingleton.WeaponLevels[geom2])
					{
						return true;
					}
				}

				// create an explosion and knockback effect
				if (weaponGenre.Equals(Weapon.weaponType.ROCKET))
				{
					AudioManager.getSingleton.playSoundFXs("explosion");

					/// push a body the projectile collides with
					g2.Body.ApplyImpulse(_physicsBody.Body.Mass * KnockBackAcceleration);

					// draw particle effects
					if (_explosionParticles[_explosionIndex] == null)
					{
						_explosionParticles[_explosionIndex] =
							ParticleManager.getSingleton.load("explosion",
							"assets/particles/explosionEffect");
						_particleLayer.AddParticle("explosion" + _explosionIndex,
							_explosionParticles[_explosionIndex]);
					}
					foreach (PointEmitter pEmitter in _explosionParticles[_explosionIndex].Emitters)
					{
						pEmitter.EmitterPosition = g1.Body.Position;
						pEmitter.IsActive = true;
					}
					if (_explosionIndex < _NUMOFEXPLOSIONS - 1)
					{
						_explosionIndex++;
					}
					else
					{
						_explosionIndex = 0;
					}
				}

				// if this is a rocket projectile, create an explosion
				if (weaponGenre.Equals(Weapon.weaponType.ROCKET))
				{
					ExplosionManager.getSingleton.createExplosion(g1.Position,
						ROCKET_SPLASH, this.projectileAnythingCollisionHandler);
				}

				deactivateProjectile();
			}

			return true;
		}

		/// <summary>
		/// Deactivates the projectile after it hit something. 
		/// </summary>
		public void deactivateProjectile()
		{
			Active = false;
			_physicsBody.Body.Position = OFF_SCREEN_POSITION;
			foreach (Geom geom in _physicsBody.GeomList)
			{
				geom.Tag = _name;
				geom.CollisionHandler = null;
			}
			_name = _name.Substring(0, _name.LastIndexOf(' '));
		}

		#endregion
	}
}
