using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using FarseerGames.FarseerPhysics.Dynamics;
using FarseerGames.FarseerPhysics.Collisions;

using rhodes.engine;
using rhodes.graphics;
using rhodes.physics;

/*
 * Weapon.cs
 * Abstract Weapon Class
 * Created By Andrew Kane
 * Modified & Updated by Rob Yates
 */

namespace sedohr.weapon
{
    /// <summary>
    /// Class representing a weapon where all weapons will derive from.
    /// </summary>
    abstract class Weapon
	{
		#region Enumerations

		/// <summary>
		/// Enumeration for the various types of weeapons we have
		/// </summary>
		public enum weaponType { NONE, PISTOL, LASER, ROCKET };

		#endregion

		#region Attributes

		/// <summary>
		/// The amount of time (in seconds) that the weapon needs to cool down for
		/// </summary>
		protected float cooldownAmount;

		/// <summary>
		/// The amount of time (in seconds) the weapon can be fire continously
		/// </summary>
		protected float weaponDuration;

		/// <summary>
		/// The type of weapon this weapon is
		/// </summary>
		protected weaponType weaponGenre;

		/// <summary>
		/// The player that has this weapon
		/// </summary>
		protected PlayerIndex playerNum;

		/// <summary>
		/// The sprite image that represents this player's weapon
		/// </summary>
		protected Sprite weaponSprite;

		/// <summary>
		/// Obtains the name of the weapon
		/// </summary>
		protected string weaponName;

		/// <summary>
		/// If this button's assigned button(s) were just pressed or not
		/// </summary>
		protected bool previouslyPressingWeaponButton = false;

		#endregion

		#region Accessors

		/// <summary>
		/// Obtains the cool down amount for a weapon
		/// </summary>
		public float cooldown
		{
			get { return cooldownAmount; }
		}

        /// <summary>
        /// Gets the weapon duration amount
        /// </summary>
        public float WeaponDuration
        {
            get { return weaponDuration; }
        }

		/// <summary>
		/// Allows us to obtain the type of the weapon
		/// </summary>
		public weaponType Genre
		{
			get { return weaponGenre; }
		}

		/// <summary>
		/// Gets or sets the visibility of the weapon
		/// </summary>
		public bool Visible
		{
			get { return weaponSprite.Visible; }
			set
			{
				weaponSprite.Visible = value;
			}
		}

		/// <summary>
		/// The sprite image that represents this player's weapon
		/// </summary>
		public Sprite Sprite
		{
			get { return weaponSprite; }
		}

		/// <summary>
		/// Obtains the name of the weapon
		/// </summary>
		public string Name
		{
			get { return weaponName; }
		}

		/// <summary>
		/// If the weapon is firing
		/// </summary>
		public bool isFiring
		{
			get { return previouslyPressingWeaponButton; }
		}

		/// <summary>
		/// Weapon constructor
		/// </summary>
		/// <param name="playerIndex">
		/// The index of the player firing this weapon
		/// </param>
        public Weapon(PlayerIndex playerIndex)
        {
			cooldownAmount = 0;
            weaponDuration = 0.0f;
            playerNum = playerIndex;
		}

		#endregion

		#region Abstract Functions

		///// <summary>
		///// Update the weapon
		///// </summary>
		///// <param name="elapsedTime">The time since the last update</param>
		///// <param name="press">
		///// True, if the fire button is being pressed. False, if not.
		///// </param>
		///// <param name="position">The weapons's position</param>
		///// <param name="rotation">The weapons's rotation</param>
		abstract public void update(float elapsedTime, bool press, Vector2 position, float rotation);

		#endregion
	}
}
